from multiprocessing import get_context
from multiprocessing.pool import Pool
import numpy as np
from .fbp_base import BackprojectorBase


# See notes in MLEMReconstructor.__init__
cct = None

try:
    # importing cupy normally won't mess with multiprocessing
    # (i.e it won't create a unsollicited context before fork())
    import cupy

    __has_cupy__ = True
except:
    __has_cupy__ = False


class CCTMLEMReconstructor:
    """
    A reconstructor for MLEM reconstruction using the CorrCT toolbox.

    Parameters
    ----------
    data_vwu_shape : tuple
        Shape of the input data, expected to be (n_slices, n_angles, n_dets). Raises an error if the shape is not 3D.
    angles_rad : numpy.ndarray
        Angles in radians for the projections. Must match the second dimension of `data_vwu_shape`.
    shifts_vu : numpy.ndarray, optional.
        Shifts in the v and u directions for each angle. If provided, must have the same number of cols as `angles_rad`. Each col is (tv,tu)
    cor : float, optional
        Center of rotation, which will be adjusted based on the sinogram width.
    n_iterations : int, optional
        Number of iterations for the MLEM algorithm. Default is 50.
    extra_options : dict, optional
        Additional options for the reconstruction process. Default options include:
        - scale_factor (float, default is 1.0): Scale factor for the reconstruction.
        - compute_shifts (boolean, default is False): Whether to compute shifts.
        - tomo_consistency (boolean, default is False): Whether to enforce tomographic consistency.
        - v_min_for_v_shifts (number, default is 0): Minimum value for vertical shifts.
        - v_max_for_v_shifts (number, default is None): Maximum value for vertical shifts.
        - v_min_for_u_shifts (number, default is 0): Minimum value for horizontal shifts.
        - v_max_for_u_shifts (number, default is None): Maximum value for horizontal shifts.
    """

    implementation = "corrct"
    default_extra_options = {
        "compute_shifts": False,
        "tomo_consistency": False,
        "v_min_for_v_shifts": 0,
        "v_max_for_v_shifts": None,
        "v_min_for_u_shifts": 0,
        "v_max_for_u_shifts": None,
        "scale_factor": 1.0,
    }

    def __init__(
        self,
        data_vwu_shape,
        angles_rad,
        shifts_uv=None,
        cor=None,  # absolute
        n_iterations=50,
        extra_options=None,
    ):

        # Do it here, since importing *anything* from corrct will spawn a Cuda context.
        # This will crash if called from within a subprocess without clean context management
        global cct  # noqa: PLW0603
        import corrct as cct

        #
        self.angles_rad = angles_rad
        self.n_iterations = n_iterations
        self.scale_factor = extra_options.get("scale_factor", 1.0)

        self._configure_extra_options(extra_options)
        self._set_sino_shape(data_vwu_shape)
        self._set_shifts(shifts_uv, cor)

    def _configure_extra_options(self, extra_options):
        self.extra_options = self.default_extra_options.copy()
        self.extra_options.update(extra_options or {})

    def _set_sino_shape(self, sinos_shape):
        if len(sinos_shape) != 3:
            raise ValueError("Expected a 3D shape")
        self.sinos_shape = sinos_shape
        self.n_sinos, self.n_angles, self.prj_width = sinos_shape
        if self.n_angles != len(self.angles_rad):
            raise ValueError(
                f"Number of angles ({len(self.angles_rad)}) does not match size of sinograms ({self.n_angles})."
            )

    def _set_shifts(self, shifts_uv, cor):
        if shifts_uv is None:
            self.shifts_vu = None
        else:
            if shifts_uv.shape[0] != self.n_angles:
                raise ValueError(
                    f"Number of shifts given ({shifts_uv.shape[0]}) does not mathc the number of projections ({self.n_angles})."
                )
            self.shifts_vu = -shifts_uv.copy().T[::-1]
        if cor is None:
            self.cor = 0.0
        else:
            self.cor = (
                -cor + (self.sinos_shape[-1] - 1) / 2.0
            )  # convert absolute to relative in the ASTRA convention, which is opposite to Nabu relative convention.

    def reset_rot_center(self, cor):
        self.cor = -cor + (self.sinos_shape[-1] - 1) / 2.0

    def reconstruct(self, data_vwu, x0=None):
        """
        data_align_vwu: numpy.ndarray or cupy array
            Raw data, with shape (n_sinograms, n_angles, width)
        output: optional
            Output array. If not provided, a new numpy array is returned
        """
        if not isinstance(data_vwu, np.ndarray):
            data_vwu = data_vwu.get()
        # data_vwu /= data_vwu.mean()

        # MLEM recons
        self.vol_geom_align = cct.models.VolumeGeometry.get_default_from_data(data_vwu)
        if self.shifts_vu is not None:
            self.prj_geom_align = cct.models.ProjectionGeometry.get_default_parallel()
            # Vertical shifts were handled in pipeline. Set them to ZERO
            self.shifts_vu[:, 0] = 0.0
            self.prj_geom_align.set_detector_shifts_vu(self.shifts_vu, self.cor)
        else:
            self.prj_geom_align = None

        variances_align = cct.processing.compute_variance_poisson(data_vwu)
        self.weights_align = cct.processing.compute_variance_weight(variances_align, normalized=True)  # , use_std=True
        self.data_term_align = cct.data_terms.DataFidelity_wl2(self.weights_align)
        solver = cct.solvers.MLEM(verbose=True, data_term=self.data_term_align)
        self.solver_opts = dict(lower_limit=0)  # , x_mask=cct.processing.circular_mask(vol_geom_align.shape_xyz[:-2])

        with cct.projectors.ProjectorUncorrected(
            self.vol_geom_align, self.angles_rad, rot_axis_shift_pix=self.cor, prj_geom=self.prj_geom_align
        ) as A:
            rec, _ = solver(A, data_vwu, iterations=self.n_iterations, x0=x0, **self.solver_opts)
        return rec * self.scale_factor


# COMPAT.
MLEMReconstructor = CCTMLEMReconstructor
# -


def _has_corrct(x):
    # should be run from within a Process
    try:
        # ruff: noqa: F401
        import corrct

        avail = True
    except (ImportError, RuntimeError, OSError, NameError):
        avail = False
    return avail


def has_corrct(safe=True):
    """
    corrct will create a cuda context upon importing any module.
    On the other hand, python will crash if creating a subprocess when cuda is already initialized.
    So to check whether corrct is available, we do the "try import" in a subprocess
    """
    if not safe:
        return _has_corrct(None)
    try:
        ctx = get_context("spawn")
        with Pool(1, context=ctx) as p:
            v = p.map(_has_corrct, [1])[0]
    except AssertionError:
        # Can get AssertionError: daemonic processes are not allowed to have children
        # if the calling code is already a subprocess
        return _has_corrct(None)
    return v


class NabuMLEMReconstructor:

    implementation = "nabu"
    default_extra_options = BackprojectorBase.default_extra_options.copy()
    default_extra_options.update(
        {
            "reg_weight": 0.0,
        }
    )

    def __init__(
        self,
        sino_shape,
        slice_shape=None,
        angles=None,
        rot_center=None,
        halftomo=False,
        extra_options=None,
        cuda_options=None,
    ):
        """
        Initialize a MLEM reconstructor.

        Parameters
        ----------
        sino_shape: tuple of int
            sinogram shape, in the form (n_angles, horiz_width)
        slice_shape: tuple of int, optional
            Slice shape, in the form (n_y, n_x)
        angles: array of float, optional
            Rotation angles in radians. If not provided, will be equispaced between [0, pi[ (or [0, 2pi[ for halftomo).
        rot_center: float, optional
            Rotation axis position, starting from the left-most pixel.
            Coordinates are set on the middle of pixels, meaning that the left side of the left-most pixel has coordinate -0.5.
        halftomo: bool, optional
            Whether to reconstruct in half-tomography (extended field of view) mode


        Other parameters
        -----------------
        extra_options: dict, optional
            Dictionary of extra options. Available are:
                - reg_weight: float
                    If set to a positive value, will enable median-regularization

        Notes
        -----
          - When reconstructing "local tomography data", it's usually best to use slice_shape to a slightly bigger extent.
          - Median regularization is based on this paper: https://link.springer.com/article/10.1007/BF01728761#preview
        """

        self._configure_extra_options(extra_options)
        if not (__has_cupy__):
            raise ImportError("Need cupy to use this class")

        # do it here to avoid creating cuda context at module import
        from cupy import ElementwiseKernel
        from cupyx.scipy.ndimage import median_filter as cupy_median_filter
        from .fbp import Backprojector, CudaClipCircle
        from .projection import Projector

        self.backproj = Backprojector(
            sino_shape,
            slice_shape=slice_shape,
            angles=angles,
            rot_center=rot_center,
            halftomo=halftomo,
            filter_name="none",
            extra_options=self.extra_options,
            backend_options=cuda_options,
        )
        self.slice_shape = self.backproj.slice_shape
        dwidth = sino_shape[-1]
        self.proj = Projector(
            self.slice_shape,
            self.backproj.angles,
            rot_center=rot_center,
            detector_width=dwidth,
            normalize=False,
            extra_options=self.extra_options,
            cuda_options=cuda_options,
        )

        # does: proj_inv = data / (|proj(x)| > eps or 1)
        self._update_projection_kernel = ElementwiseKernel(
            "float32 proj, float32 proj_data, float32 eps",
            "float32 proj_inv",
            "proj_inv = proj_data / ((fabsf(proj) > eps) ? (proj) : (1.0f))",
            name="update_projection",
        )
        if self._do_regularization:
            # does: x_k /= B(1) + Lambda * ((x_k - m)*m)/(m^2 + eps)
            self._regularize_kernel = ElementwiseKernel(
                "float32 x_med, float32 pto, float32 eps, float32 lambda",
                "float32 x",
                # x /= pto + lambda * (x - x_med)*x_med / (x_med * x_med + eps) ;
                """
                float oinv = 1.0f/(pto + lambda * (x - x_med)*x_med / (x_med * x_med + eps));
                if (oinv < 0.0f) oinv = 0.0f;
                x *= oinv;
                """,
                name="regularize_reconstruction",
            )
            # median filter
            self.median_filter = cupy_median_filter

        if self._do_clip_circle:
            self.clip_circle = CudaClipCircle(self.slice_shape, outer_circle_value=self._clip_circle_val)

        cuda = self.proj.cuda_processing
        self.cuda = cuda
        ones = np.ones(sino_shape, "f")
        self._pto = cuda.to_device("pto", self.backproj.backproj(ones)).astype("f")
        self._x = cuda.allocate_array("x", self._pto.shape)
        self._y = cuda.allocate_array("y", self._x.shape)
        self._proj = cuda.allocate_array("proj", sino_shape)
        self._proj_inv = cuda.allocate_array("proj_inv", sino_shape)
        if self._do_regularization:
            self._x_med = cuda.allocate_array("x_med", self._x.shape)

    def _configure_extra_options(self, extra_options):
        self.extra_options = self.default_extra_options.copy()
        self.extra_options.update(extra_options or {})
        # Always use "centered_axis" for iterative reconstruction
        self.extra_options["centered_axis"] = True

        # Usually we don't want to clip the outer circle at each iteration. Do it only at the end.
        # pop() will remove from dict, thus disable clip_circle in backprojector
        self._do_clip_circle = self.extra_options.pop("clip_outer_circle", False)
        self._clip_circle_val = self.extra_options.get("outer_circle_value", 0)

        # These parameters are not needed for Backprojector and Projector
        self.extra_options.pop("padding_mode", None)
        self._scale_factor = self.extra_options.pop("scale_factor", None)
        #
        self._reg_weight = self.extra_options.get("reg_weight", 0.0)
        self._do_regularization = self._reg_weight > 0

    def reset_rot_center(self, cor):
        self.backproj.reset_rot_center(cor)
        self.proj.reset_rot_center(cor)

    def reconstruct(self, sino, n_iterations=10, output=None, eps=1e-6, regularization_weight=None):
        if isinstance(sino, np.ndarray):
            sino = self.cuda.to_device("sino", sino)

        # data should not contain negative values
        cupy.clip(sino, eps, None, out=sino)

        x = self._x
        y = self._y
        proj = self._proj
        proj_inv = self._proj_inv

        if output is not None:
            self.cuda.check_array(output, x.shape)
            x = output

        if self._do_regularization:
            x_med = self._x_med
            reg_weight = (regularization_weight or self._reg_weight) * self.backproj.n_angles
        else:
            oinv = self._pto.copy()
            cupy.divide(1, oinv, out=oinv)

        x.fill(1)  # update is multiplicative, start with ones (not zeros)
        # Not sure about that - requested by a beamline, that's too specific
        if self._do_clip_circle and self._do_regularization:
            self.clip_circle(x)
        #

        for k in range(n_iterations):
            # proj = P(x)
            self.proj.projection(x, output=proj)

            # proj_inv = data / (|proj(x)| > eps or 1)
            self._update_projection_kernel(proj, sino, eps, proj_inv)

            if self._do_regularization:
                # x /= B(1) + Lambda * (x - m)*m / (m^2 + eps)
                self.median_filter(x, size=(3, 3), output=x_med)
                self._regularize_kernel(x_med, self._pto, eps, reg_weight, x)

            else:
                # x /= B(1)
                x *= oinv

            # x *= B(proj_inv)
            self.backproj.backproj(proj_inv, output=y)
            x *= y

        if self._scale_factor is not None:
            x *= self._scale_factor

        if self._do_clip_circle:
            self.clip_circle(x)  # TODO different radius if regularization ?

        return x

    __call__ = reconstruct



import numpy as np
from numpy.polynomial.polynomial import Polynomial, polyval
from scipy import fft

from ..utils import calc_padding_lengths, deprecation_warning
from .alignment import AlignmentBase, plt


class DetectorTranslationAlongBeam(AlignmentBase):
    def find_shift(
        self,
        img_stack: np.ndarray,
        img_pos: np.array,
        roi_yxhw=None,
        median_filt_shape=None,
        padding_mode=None,
        peak_fit_radius=1,
        high_pass=None,
        low_pass=None,
        return_shifts=False,
        use_adjacent_imgs=False,
    ):
        """Find the vertical and horizontal shifts for translations of the
        detector along the beam direction.

        These shifts are in pixels-per-unit-translation, and they are due to
        the misalignment of the translation stage, with respect to the beam
        propagation direction.

        To compute the vertical and horizontal tilt angles from the obtained `shift_pix`:

        >>> tilt_deg = np.rad2deg(np.arctan(shift_pix * pixel_size))

        where `pixel_size` and and the input parameter `img_pos` have to be
        expressed in the same units.

        Parameters
        ----------
        img_stack: numpy.ndarray
            A stack of images (usually 4) at different distances
        img_pos: numpy.ndarray
            Position of the images along the translation axis
        roi_yxhw: (2, ) or (4, ) numpy.ndarray, tuple, or array, optional
            4 elements vector containing: vertical and horizontal coordinates
            of first pixel, plus height and width of the Region of Interest (RoI).
            Or a 2 elements vector containing: plus height and width of the
            centered Region of Interest (RoI).
            Default is None -> deactivated.
        median_filt_shape: (2, ) numpy.ndarray, tuple, or array, optional
            Shape of the median filter window. Default is None -> deactivated.
        padding_mode: str in numpy.pad's mode list, optional
            Padding mode, which determines the type of convolution. If None or
            'wrap' are passed, this resorts to the traditional circular convolution.
            If 'edge' or 'constant' are passed, it results in a linear convolution.
            Default is the circular convolution.
            All options are:
                None | 'constant' | 'edge' | 'linear_ramp' | 'maximum' | 'mean'
                | 'median' | 'minimum' | 'reflect' | 'symmetric' |'wrap'
        peak_fit_radius: int, optional
            Radius size around the max correlation pixel, for sub-pixel fitting.
            Minimum and default value is 1.
        low_pass: float or sequence of two floats
            Low-pass filter properties, as described in `nabu.misc.fourier_filters`.
        high_pass: float or sequence of two floats
            High-pass filter properties, as described in `nabu.misc.fourier_filters`.
        return_shifts: boolean, optional
            Adds a third returned argument, containing the pixel shifts of each
            image with respect to the first one in the stack. Defaults to False.
        use_adjacent_imgs: boolean, optional
            Compute correlation between adjacent images.
            It can be used when dealing with large shifts, to avoid overflowing the shift.
            This option allows to replicate the behavior of the reference function `alignxc.m`
            However, it is detrimental to shift fitting accuracy. Defaults to False.

        Returns
        -------
        coeff_v: float
            Estimated vertical shift in pixel per unit-distance of the detector translation.
        coeff_h: float
            Estimated horizontal shift in pixel per unit-distance of the detector translation.
        shifts_vh: list, optional
            The pixel shifts of each image with respect to the first image in the stack.
            Activated if return_shifts is True.

        Examples
        --------
        The following example creates a stack of shifted images, and retrieves the computed shift.
        Here we use a high-pass filter, due to the presence of some low-frequency noise component.

        >>> import numpy as np
        ... import scipy as sp
        ... import scipy.ndimage
        ... from nabu.preproc.alignment import  DetectorTranslationAlongBeam
        ...
        ... tr_calc = DetectorTranslationAlongBeam()
        ...
        ... stack = np.zeros([4, 512, 512])
        ...
        ... # Add low frequency spurious component
        ... for i in range(4):
        ...     stack[i, 200 - i * 10, 200 - i * 10] = 1
        ... stack = sp.ndimage.filters.gaussian_filter(stack, [0, 10, 10.0]) * 100
        ...
        ... # Add the feature
        ... x, y = np.meshgrid(np.arange(stack.shape[-1]), np.arange(stack.shape[-2]))
        ... for i in range(4):
        ...     xc = x - (250 + i * 1.234)
        ...     yc = y - (250 + i * 1.234 * 2)
        ...     stack[i] += np.exp(-(xc * xc + yc * yc) * 0.5)
        ...
        ... # Image translation along the beam
        ... img_pos = np.arange(4)
        ...
        ... # Find the shifts from the features
        ... shifts_v, shifts_h = tr_calc.find_shift(stack, img_pos, high_pass=1.0)
        ... print(shifts_v, shifts_h)
        >>> ( -2.47 , -1.236 )

        and the following commands convert the shifts in angular tilts:

        >>> tilt_v_deg = np.rad2deg(np.arctan(shifts_v * pixel_size))
        >>> tilt_h_deg = np.rad2deg(np.arctan(shifts_h * pixel_size))

        To enable the legacy behavior of `alignxc.m` (correlation between adjacent images):

        >>> shifts_v, shifts_h = tr_calc.find_shift(stack, img_pos, use_adjacent_imgs=True)

        To plot the correlation shifts and the fitted straight lines for both directions:

        >>> tr_calc = DetectorTranslationAlongBeam(verbose=True)
        ... shifts_v, shifts_h = tr_calc.find_shift(stack, img_pos)
        """
        self._check_img_stack_size(img_stack, img_pos)

        if peak_fit_radius < 1:
            self.logger.warning("Parameter peak_fit_radius should be at least 1, given: %d instead." % peak_fit_radius)
            peak_fit_radius = 1

        num_imgs = img_stack.shape[0]
        img_shape = img_stack.shape[-2:]
        roi_yxhw = self._determine_roi(img_shape, roi_yxhw)

        img_stack = self._prepare_image(img_stack, roi_yxhw=roi_yxhw, median_filt_shape=median_filt_shape)

        # do correlations
        ccs = [
            self._compute_correlation_fft(
                img_stack[ii - 1 if use_adjacent_imgs else 0, ...],
                img_stack[ii, ...],
                padding_mode,
                high_pass=high_pass,
                low_pass=low_pass,
            )
            for ii in range(1, num_imgs)
        ]

        img_shape = ccs[0].shape  # cc.shape can differ from img.shape, e.g. in case of odd number of cols.
        cc_vs = np.fft.fftfreq(img_shape[-2], 1 / img_shape[-2])
        cc_hs = np.fft.fftfreq(img_shape[-1], 1 / img_shape[-1])

        shifts_vh = np.zeros((num_imgs, 2))
        for ii, cc in enumerate(ccs):
            (f_vals, fv, fh) = self.extract_peak_region_2d(cc, peak_radius=peak_fit_radius, cc_vs=cc_vs, cc_hs=cc_hs)
            shifts_vh[ii + 1, :] = self.refine_max_position_2d(f_vals, fv, fh)

        if use_adjacent_imgs:
            shifts_vh = np.cumsum(shifts_vh, axis=0)

        # Polynomial.fit is supposed to be more numerically stable than polyfit
        # (according to numpy)
        coeffs_v = Polynomial.fit(img_pos, shifts_vh[:, 0], deg=1).convert().coef
        coeffs_h = Polynomial.fit(img_pos, shifts_vh[:, 1], deg=1).convert().coef
        # In some cases (singular matrix ?) the output is [0] while in some other its [eps, eps].
        if len(coeffs_v) == 1:
            coeffs_v = np.array([coeffs_v[0], coeffs_v[0]])
        if len(coeffs_h) == 1:
            coeffs_h = np.array([coeffs_h[0], coeffs_h[0]])

        if self.verbose:
            self.logger.info(
                "Fitted pixel shifts per unit-length: vertical = %f, horizontal = %f" % (coeffs_v[1], coeffs_h[1])
            )
            f, axs = plt.subplots(1, 2)
            self._add_plot_window(f, ax=axs)
            axs[0].scatter(img_pos, shifts_vh[:, 0])
            axs[0].plot(img_pos, polyval(img_pos, coeffs_v), "-C1")
            axs[0].set_title("Vertical shifts")
            axs[1].scatter(img_pos, shifts_vh[:, 1])
            axs[1].plot(img_pos, polyval(img_pos, coeffs_h), "-C1")
            axs[1].set_title("Horizontal shifts")
            plt.show(block=False)

        if return_shifts:
            return coeffs_v[1], coeffs_h[1], shifts_vh
        else:
            return coeffs_v[1], coeffs_h[1]


class OppositeImages2DRegistrationOctaveAccurate(AlignmentBase):
    """This is a Python implementation of Octave/fastomo3/accurate COR estimator.
    The Octave 'accurate' function is renamed `local_correlation`.
    The Nabu standard `find_shift` has the same API as the other COR estimators (sliding, growing...)
    Note that this estimator returns a 2d estimation of the shift between the 2 input images.
    If you need only the COR estimation, see estimation.cor.CenterOfRotationOctaveAccurate.
    """

    def _cut(self, im, nrows, ncols, new_center_row=None, new_center_col=None):
        """Cuts a sub-matrix out of a larger matrix.
        Cuts in the center of the original matrix, except if new center is specified
        NO CHECKING of validity indices sub-matrix!

        Parameters
        ----------
        im : array.
            Original matrix
        nrows : int
            Number of rows in the output matrix.
        ncols : int
            Number of columns in the output matrix.
        new_center_row : int
            Index of center row around which to cut (default: None, i.e. center)
        new_center_col : int
            Index of center column around which to cut (default: None, i.e. center)

        Returns
        -------
        nrows x ncols array.

        Examples
        --------
        im_roi = cut(im, 1024, 1024)                -> cut center 1024x1024 pixels
        im_roi = cut(im, 1024, 1024, 600.5, 700.5)  -> cut 1024x1024 pixels around pixels (600-601, 700-701)

        Author: P. Cloetens <cloetens@esrf.eu>
        2023-11-06 J. Lesaint <jerome.lesaint@esrf.fr>

        * See octave-archive for the original Octave code.
        * 2023-11-06: Python implementation. Comparison seems OK.
        """
        [n, m] = im.shape
        if new_center_row is None:
            new_center_row = (n + 1) / 2
        if new_center_col is None:
            new_center_col = (m + 1) / 2

        rb = int(np.round(0.5 + new_center_row - nrows / 2))
        rb = int(np.round(new_center_row - nrows / 2))
        re = int(nrows + rb)
        cb = int(np.round(0.5 + new_center_col - ncols / 2))
        cb = int(np.round(new_center_col - ncols / 2))
        ce = int(ncols + cb)

        return im[rb:re, cb:ce]

    def _checkifpart(self, rapp, rapp_hist):
        res = 0
        for k in range(rapp_hist.shape[0]):
            if np.allclose(rapp, rapp_hist[k, :]):
                res = 1
                return res
        return res

    def _interpolate(self, input_, shift, mode="mean", interpolation_method="linear"):
        """Applies to the input a translation by a vector `shift`. Based on
        `scipy.ndimage.affine_transform` function.
        JL: This Octave function was initially used in the refine clause of the local_correlation (Octave find_shift).
        Since find_shift is always called with refine=False in Octave, refine is not implemented (see local_interpolation())
        and this function becomes useless.

        Parameters
        ----------
        input_ : array
            Array to which the translation is applied.
        shift : tuple, list or array of length 2.
        mode : str
            Type of padding applied to the unapplicable areas of the output image.
            Default `mean` is a constant padding with the mean of the input array.
            `mode` must belong to 'reflect', 'grid-mirror', 'constant', 'grid-constant', 'nearest', 'mirror', 'grid-wrap', 'wrap'
            See `scipy.ndimage.affine_transform` for details.
        interpolation_method : str or int.
            The interpolation is based on spline interpolation.
            Either 0, 1, 2, 3, 4 or 5: order of the spline interpolation functions.
            Or one among 'linear','cubic','pchip','nearest','spline' (Octave legacy).
                'nearest' is equivalent to 0
                'linear' is equivalent to 1
                'cubic','pchip','spline' are equivalent to 3.
        """
        admissible_modes = (
            "reflect",
            "grid-mirror",
            "constant",
            "grid-constant",
            "nearest",
            "mirror",
            "grid-wrap",
            "wrap",
        )
        admissible_interpolation_methods = ("linear", "cubic", "pchip", "nearest", "spline")

        from scipy.ndimage import affine_transform

        [s0, s1] = shift
        matrix = np.zeros([2, 3], dtype=float)
        matrix[0, 0] = 1.0
        matrix[1, 1] = 1.0
        matrix[:, 2] = [-s0, -s1]  # JL: due to transf. convention diff in Octave and scipy (push fwd vs pull back)

        if interpolation_method == "nearest":
            order = 0
        elif interpolation_method == "linear":
            order = 1
        elif interpolation_method in ("pchip", "cubic", "spline"):
            order = 3
        elif interpolation_method in (0, 1, 2, 3, 4, 5):
            order = interpolation_method
        else:
            raise ValueError(
                f"Interpolation method is {interpolation_method} and should either an integer between 0 (inc.) and 5 (inc.) or in {admissible_interpolation_methods}."
            )

        if mode == "mean":
            mode = "constant"
            cval = input_.mean()
            return affine_transform(input_, matrix, mode=mode, order=order, cval=cval)
        elif mode not in admissible_modes:
            raise ValueError(f"Pad method is {mode} and should be in {admissible_modes}.")

        return affine_transform(input_, matrix, mode=mode, order=order)

    def _local_correlation(
        self,
        z1,
        z2,
        maxsize=(5, 5),
        cor_estimate=(0, 0),
        refine=None,
        pmcc=False,
        normalize=True,
    ):
        """Returns the 2D shift in pixels between two images.
        It looks for a local optimum around the initial shift cor_estimate
        and within a window 'maxsize'.
        It uses variance of the difference of the normalized images or PMCC
        It adapts the shift estimate in case optimum is at the edge of the window
        If 'maxsize' is set to 0, it will only use approximate shift (+ refine possibly)
        Set 'cor_estimate' to allow for the use of any initial shift estimation.

        When not successful (stuck in loop or edge reached), returns [nan nan]
        Positive values corresponds to moving z2 to higher values of the index
        to compensate drift: interpolate(f)(z2, row, column)

        Parameters
        ----------
        z1,z2 : 2D arrays.
            The two (sub)images to be compared.

        maxsize : 2-list. Default [5,5]
            Size of the search window.

        cor_estimate:
            Initial guess of the center of rotation.

        refine: Boolean or None (default is None)
            Wether the initial guess should be refined of not.

        pmcc: Boolean (default is False)
            Use Pearson correlation coefficient  i.o. variance.

        normalize: Boolean (default is True)
            Set mean of each image to 1 if True.

        Returns
        -------
        c = [row,column] (or [NaN,NaN] if unsuccessful.)

        2007-01-05 P. Cloetens cloetens@esrf.eu
        * Initial revision
        2023-11-10 J. Lesaint jerome.lesaint@esrf.fr
        * Python conversion.
        """

        if type(maxsize) in (float, int):
            maxsize = [int(maxsize), int(maxsize)]
        elif type(maxsize) in (tuple, list):
            maxsize = [int(maxsize[0]), int(maxsize[1])]
        elif maxsize in ([], None, ""):
            maxsize = [5, 5]

        if refine is None:
            refine = np.allclose(maxsize, 0.0)

        if normalize:
            z1 /= np.mean(z1)
            z2 /= np.mean(z2)

        #####################################
        # JL : seems useless since func is always called with a first approximate.
        ## determination of approximative shift (manually or Fourier correlation)
        # if isinstance(cor_estimate,str):
        #    if cor_estimate in ('fft','auto','fourier'):
        #        padding_mode = None
        #        cor_estimate = self._compute_correlation_fft(
        #            z1,
        #            z2,
        #            padding_mode,
        #            high_pass=self.high_pass,
        #            low_pass=self.low_pass
        #        )
        #    elif cor_estimate in ('manual','man','m'):
        #        cor_estimate = None
        #        # No ImageJ plugin here :
        #        # rapp = ij_align(z1,z2)

        ####################################
        # check if refinement with realspace correlation is required
        # otherwise keep result as it is
        if np.allclose(maxsize, 0):
            shiftfound = 1
            if refine:
                c = np.round(np.array(cor_estimate, dtype=int))
            else:
                c = np.array(cor_estimate, dtype=int)
        else:
            shiftfound = 0
            cor_estimate = np.round(np.array(cor_estimate, dtype=int))

        rapp_hist = []
        if np.sum(np.abs(cor_estimate) + 1 >= z1.shape):
            self.logger.debug(f"Approximate shift of [{cor_estimate[0]},{cor_estimate[1]}] is too large, setting [0 0]")
            cor_estimate = np.array([0, 0])
        maxsize = np.minimum(maxsize, np.floor((np.array(z1.shape) - 1) / 2)).astype(int)
        maxsize = np.minimum(maxsize, np.array(z1.shape) - np.abs(cor_estimate) - 1).astype(int)

        while not shiftfound:
            # Set z1 region
            # Rationale: the (shift[0]+maxsize[0]:,shift[1]+maxsize[1]:) block of z1 should match
            # the (maxsize[0]:,maxisze[1]:)-upper-left corner of z2.
            # We first extract this z1 block.
            # Then, take moving z2-block according to maxsize.
            # Of course, care must be taken with borders, hence the various max,min calls.

            # Extract the reference block
            shape_ar = np.array(z1.shape)
            cor_ar = np.array(cor_estimate)
            maxsize_ar = np.array(maxsize)

            z1beg = np.maximum(cor_ar + maxsize_ar, np.zeros(2, dtype=int))
            z1end = shape_ar + np.minimum(cor_ar - maxsize_ar, np.zeros(2, dtype=int))

            z1p = z1[z1beg[0] : z1end[0], z1beg[1] : z1end[1]].flatten()

            # Build local correlations array.
            window_shape = (2 * int(maxsize[0]) + 1, 2 * int(maxsize[1]) + 1)
            cc = np.zeros(window_shape)

            # Prepare second block indices
            z2beg = (cor_ar + maxsize_ar > 0) * cc.shape + (cor_ar + maxsize_ar <= 0) * (shape_ar - z1end + z1beg) - 1
            z2end = z2beg + z1end - z1beg

            if pmcc:
                std_z1p = z1p.std()
            if normalize == 2:
                z1p /= z1p.mean()

            for k in range(cc.shape[0]):
                for l in range(cc.shape[1]):  # noqa: E741
                    if pmcc:
                        z2p = z2[z2beg[0] - k : z2end[0] - k, z2beg[1] - l : z2end[1] - l].flatten()
                        std_z2p = z2p.std()
                        cc[k, l] = -np.cov(z1p, z2p, rowvar=True)[1, 0] / (std_z1p * std_z2p)
                    else:
                        if normalize == 2:
                            z2p = z2[z2beg[0] - k : z2end[0] - k, z2beg[1] - l : z2end[1] - l].flatten()
                            z2p /= z2p.mean()
                            z2p -= z1p
                        else:
                            z2p = z2[z2beg[0] - k : z2end[0] - k, z2beg[1] - l : z2end[1] - l].flatten()
                            z2p -= z1p
                        cc[k, l] = ((z2p - z2p.mean()) ** 2).sum()
                        # cc(k,l) = std(z1p./z2(z2beg(1)-k:z2end(1)-k,z2beg(2)-l:z2end(2)-l)(:));

            c = np.unravel_index(np.argmin(cc, axis=None), shape=cc.shape)

            if not np.sum((c == 0) + (c == np.array(cc.shape) - 1)):
                # check that we are not at the edge of the region that was sampled
                x = np.array([-1, 0, 1])
                tmp = self.refine_max_position_2d(
                    cc[c[0] - 1 : c[0] + 2, c[1] - 1 : c[1] + 2], x, x, logger=self.logger
                )
                c += tmp
            shiftfound = True

            c += z1beg - z2beg

            rapp_hist = []
            if not shiftfound:
                cor_estimate = c
                # Check that new shift estimate was not already done (avoid eternal loop)
                if self._checkifpart(cor_estimate, rapp_hist):
                    self.logger.debug("Stuck in loop?")
                    refine = True
                    shiftfound = True
                    c = np.array([np.nan, np.nan])
                else:
                    rapp_hist.append(cor_estimate)
                    self.logger.debug(f"Changing shift estimate: {cor_estimate}")
                    maxsize = np.minimum(maxsize, np.array(z1.shape) - np.abs(cor_estimate) - 1).astype(int)
                    if (maxsize == 0).sum():
                        self.logger.debug("Edge of image reached")
                        refine = False
                        shiftfound = True
                        c = np.array([np.nan, np.nan])
            elif len(rapp_hist) > 0:
                self.logger.debug("\n")

        ####################################
        # refine result; useful when shifts are not integer values
        # JL: I don't understand why this refine step should be useful.
        # In Octave, from fastomo.m, refine is always set to False.
        # So this could be ignored.
        # I keep it for future use if it proves useful.
        # if refine:
        #    if debug:
        #        print('Refining solution ...')
        #    z2n = self.interpolate(z2,c)
        #    indices = np.ceil(np.abs(c)).astype(int)
        #    z1p = np.roll(z1,((c>0) * (-1) * indices),[0,1])
        #    z1p = z1p[1:-indices[0]-1,1:-indices[1]-1].flatten()
        #    z2n = np.roll(z2n,((c>0) * (-1) * indices),[0,1])
        #    z2n = z2n[:-indices[0],:-indices[1]]
        #    ccrefine = np.zeros([3,3])
        #    [n2,m2] = z2n.shape
        #    for k in range(3):
        #        for l in range(3):
        #            z2p = z1p - z2n[2-k:n2-k,2-l:m2-l].flatten()
        #            ccrefine[k,l] = ((z2p - z2p.mean())**2).sum()
        #    x = np.array([-1,0,1])
        #    crefine = self.refine_max_position_2d(ccrefine, x, x)
        #    #crefine = min2par(ccrefine)

        #    # Check if the refinement is effectively confined to subpixel
        #    if (np.abs(crefine) >= 1).sum():
        #        self.logger.info("Problems refining result\n")
        #    else:
        #        c += crefine

        return c

    def find_shift(
        self,
        img_1,
        img_2,
        side="center",
        roi_yxhw=None,
        median_filt_shape=None,
        padding_mode=None,
        low_pass=0.01,
        high_pass=None,
        maxsize=(5, 5),
        refine=None,
        pmcc=False,
        normalize=True,
        limz=0.5,
        return_relative_to_middle=None,
    ):
        # COMPAT.
        if return_relative_to_middle is None:
            deprecation_warning(
                "The current default behavior is to return the shift relative the the middle of the image. In a future release, this function will return the shift relative to the left-most pixel. To keep the current behavior, please use 'return_relative_to_middle=True'.",
                do_print=True,
                func_name="CenterOfRotationOctaveAccurate.find_shift",
            )
            return_relative_to_middle = True  # the kwarg above will be False by default in a future release
        # ---
        self._check_img_pair_sizes(img_1, img_2)

        img_shape = img_2.shape
        roi_yxhw = self._determine_roi(img_shape, roi_yxhw)

        img_1 = self._prepare_image(img_1, roi_yxhw=roi_yxhw, median_filt_shape=median_filt_shape)
        img_2 = self._prepare_image(img_2, roi_yxhw=roi_yxhw, median_filt_shape=median_filt_shape)

        cc = self._compute_correlation_fft(
            img_1,
            img_2,
            padding_mode,
            high_pass=high_pass,
            low_pass=low_pass,
        )

        # We use fftshift to deal more easily with negative shifts.
        # This has a cost of subtracting half the image shape afterward.
        shift = np.unravel_index(np.argmax(np.fft.fftshift(cc)), img_shape)
        shift -= np.array(img_shape) // 2

        # The real "accurate" starts here (i.e. the octave findshift() func).
        if np.abs(shift[0]) > 10 * limz:
            # This is suspiscious. We don't trust results of correlate.
            self.logger.warning(f"Pre-correlation yields {shift[0]} pixels vertical motion")
            self.logger.warning("We do not consider it.")
            shift = (0, 0)

        # Limit the size of region for comparison to cutsize in both directions.
        # Hard-coded?
        cutsize = img_shape[1] // 2
        oldshift = np.round(shift).astype(int)
        if (img_shape[0] > cutsize) or (img_shape[1] > cutsize):
            im0 = self._cut(img_1, min(img_shape[0], cutsize), min(img_shape[1], cutsize))
            im1 = self._cut(
                np.roll(img_2, oldshift, axis=(0, 1)), min(img_shape[0], cutsize), min(img_shape[1], cutsize)
            )
            shift = oldshift + self._local_correlation(
                im0,
                im1,
                maxsize=maxsize,
                refine=refine,
                pmcc=pmcc,
                normalize=normalize,
            )
        else:
            shift = self._local_correlation(
                img_1,
                img_2,
                maxsize=maxsize,
                cor_estimate=oldshift,
                refine=refine,
                pmcc=pmcc,
                normalize=normalize,
            )
        if ((shift - oldshift) ** 2).sum() > 4:
            self.logger.warning(f"Pre-correlation ({oldshift}) and accurate correlation ({shift}) are not consistent.")
            self.logger.warning("Please check!!!")

        offset = shift[1] / 2

        if np.abs(shift[0]) > limz:
            self.logger.debug("Verify alignment or sample motion.")
            self.logger.debug(f"Verical motion: {shift[0]} pixels.")
            self.logger.debug(f"Offset?: {offset} pixels.")
        else:
            self.logger.debug(f"Offset?: {offset} pixels.")

        if not (return_relative_to_middle):
            offset += (img_shape[1] - 1) / 2

        return shift


def _fft_pad(i, axes=None, padding_mode="constant"):
    pad_len = calc_padding_lengths(i.shape, np.array(i.shape) * 2)
    i_p = np.pad(i, pad_len, mode=padding_mode)
    return fft.fftn(i_p)


def estimate_shifts(im1, im2):
    """
    Simple implementation of shift estimation between two images, based on phase cross correlation.
    """
    pr = _fft_pad(im1) * _fft_pad(im2).conjugate()
    pr_n = pr / np.maximum(1e-7, np.abs(pr))
    corr = np.fft.fftshift(fft.ifftn(pr_n).real)
    argmax = np.array(np.unravel_index(np.argmax(corr), pr.shape))
    shp = np.array(pr.shape)
    argmax_refined = refine_parabola_2D(corr, argmax)
    argmax = argmax + argmax_refined
    return shp // 2 - np.array(argmax)


def refine_parabola_2D(im_vals, argmax):
    argmax = tuple(argmax)
    maxval = im_vals[argmax]
    ny, nx = im_vals.shape

    iy, ix = np.array(argmax, dtype=np.intp)
    ixm, ixp = (ix - 1) % nx, (ix + 1) % nx
    iym, iyp = (iy - 1) % ny, (iy + 1) % ny

    F = maxval
    A = (im_vals[iy, ixp] + im_vals[iy, ixm]) / 2 - F
    D = (im_vals[iy, ixp] - im_vals[iy, ixm]) / 2
    B = (im_vals[iyp, ix] + im_vals[iym, ix]) / 2 - F
    E = (im_vals[iyp, ix] - im_vals[iym, ix]) / 2
    C = (im_vals[iyp, ixp] - im_vals[iym, ixp] - im_vals[iyp, ixm] + im_vals[iym, ixm]) / 4
    det = C**2 - 4 * A * B
    dx = (2 * B * D - C * E) / det
    dy = (2 * A * E - C * D) / det
    dx = np.clip(dx, -0.5, 0.5)
    dy = np.clip(dy, -0.5, 0.5)
    return (dy, dx)



import os
from math import ceil
import tempfile
from warnings import warn
import numpy as np
from tomoscan.esrf.volume import TIFFVolume, JP2KVolume, MultiTIFFVolume, RawVolume, HDF5Volume, EDFVolume
from tomoscan.esrf.volume.singleframebase import VolumeSingleFrameBase
from tomoscan.io import HDF5File
from ..utils import check_supported
from ..resources.utils import get_available_ram_GB
from ..io.utils import get_first_hdf5_entry
from .utils import parse_params_values
from .cli_configs import FlipVolumeVerticallyConfig


def _infer_h5_data_path(file_path):
    entry = get_first_hdf5_entry(file_path)
    with HDF5File(file_path, "r") as f:
        if "reconstruction" in f[entry]:
            # Assume the file was created by nabu
            res = f"{entry}/reconstruction"
        else:
            # Assume NXTomo file
            res = f"{entry}/instrument/detector/data"
    return res


def get_volume_reader(volume_path, volume_type, file_prefix, h5_data_path=None):
    volume_readers = {
        "tiff3d": MultiTIFFVolume,
        "tiff": TIFFVolume,
        "hdf5": HDF5Volume,
        "h5": HDF5Volume,  # alias
        "edf": EDFVolume,
        "jp2": JP2KVolume,
        "jpeg2000": JP2KVolume,  # alias
        "vol": RawVolume,
        "raw": RawVolume,  # alias
    }
    check_supported(volume_type, list(volume_readers.keys()), "volume type")
    reader_cls = volume_readers[volume_type]

    if issubclass(reader_cls, VolumeSingleFrameBase):
        if file_prefix in [None, ""]:
            raise ValueError(f"Need 'file_prefix' volume type '{volume_type}'")
        reader_init_kwargs = {
            "folder": volume_path,
            "volume_basename": file_prefix,
        }
    else:
        reader_init_kwargs = {"file_path": volume_path}
        if reader_cls == HDF5Volume:
            h5_data_path = h5_data_path or _infer_h5_data_path(volume_path)
            reader_init_kwargs["data_path"] = h5_data_path
    reader_init_kwargs["overwrite"] = True
    reader = reader_cls(**reader_init_kwargs)
    return reader


def flip_singleframe_volume_vertically(volume):
    """
    Flip up<->down the volume for "single-frame" volumes (tiff, edf, jpeg2000)
    """
    files = list(volume.browse_data_files())
    n_files = len(files)
    scheme = volume.data_url.scheme()
    for i in range(n_files // 2):
        data_up = volume.read_file(files[i])
        data_down = volume.read_file(files[n_files - 1 - i])
        volume.save_frame(data_down, files[i], scheme)
        volume.save_frame(data_up, files[n_files - 1 - i], scheme)


def _get_tmpdir_capacity_GB(tmp_dir):
    stat = os.statvfs(tmp_dir)
    tmp_dir_capacity_available_GB = stat.f_frsize * stat.f_bavail / 1e9
    return tmp_dir_capacity_available_GB


def flip_multiframe_volume_vertically(volume, mem_fraction=0.25):
    """
    Flip up<->down the volume for "multi-frame" volumes (hdf5, tiff3d, raw)
    """

    vol_shape = volume.get_volume_shape()
    vol_nbytes_GB = np.prod(vol_shape) * 4 / 1e9  # assume 4 bytes/voxel
    mem_used_GB = mem_fraction * get_available_ram_GB()

    if vol_nbytes_GB < mem_used_GB:
        # We can use the naive approach: load everything in RAM, flip, write
        volume.load_data()
        volume.data = volume.data[::-1, :, :]
        volume.save_data()

    else:
        # Load/write by chunks.
        # Use numpy memory maps as a buffer
        tmp_dir = tempfile.gettempdir()
        tmp_dir_capacity_available_GB = _get_tmpdir_capacity_GB(tmp_dir)
        if tmp_dir_capacity_available_GB < vol_nbytes_GB:
            raise RuntimeError("Not enough RAM nor tmp space to invert the volume with current implementation")
        tmpfile = os.path.join(tmp_dir, os.path.basename(volume.file_path) + ".swp")
        warn(f"Not enough RAM to flip volume in-place, using temporary file {tmpfile}", RuntimeWarning)

        def _get_mmap(mode):
            return np.memmap(
                tmpfile,
                dtype=np.float32,
                mode=mode,
                offset=0,
                shape=vol_shape,
            )

        n_chunks = ceil(vol_nbytes_GB / mem_used_GB)
        chunk_size = vol_shape[0] // n_chunks
        try:
            buffer = _get_mmap("w+")
            for i in range(n_chunks):
                i_start = i * chunk_size
                i_stop = min((i + 1) * chunk_size, vol_shape[0])
                buffer[i_start:i_stop] = volume.load_chunk((slice(i_start, i_stop), slice(None), slice(None)))
            buffer.flush()
            del buffer
            buffer = _get_mmap("r")
            volume.data = buffer[::-1]
            volume.save_data()
        finally:
            os.remove(tmpfile)


def flip_volume_vertically(volume_path, volume_type, file_prefix, h5_data_path=None, mem_fraction=0.25):
    mem_fraction = np.clip(mem_fraction, 0.0, 1.0)
    volume = get_volume_reader(volume_path, volume_type, file_prefix, h5_data_path=h5_data_path)
    if isinstance(volume, VolumeSingleFrameBase):
        flip_singleframe_volume_vertically(volume)
    else:
        flip_multiframe_volume_vertically(volume, mem_fraction=mem_fraction)


def flip_volume_vertically_app():
    args = parse_params_values(FlipVolumeVerticallyConfig, parser_description="Flip a volume upside-down, in-place")
    flip_volume_vertically(args["input"], args["vol_type"].lower(), args["file_prefix"])
    exit(0)



from math import sqrt, ceil
import numpy as np

try:
    import matplotlib.pyplot as plt
except ImportError:
    plt = None


def clip_std(arr, n=3):
    m = np.mean(arr)
    s = np.std(arr)
    return np.clip(arr, m - n * s, m + n * s)


def ims(img, cmap="gray", legend=None, colorbar=True, share=True, rescale_n_std=None):
    """
    image visualization utility.

    img: 2D numpy.ndarray, or list of 2D numpy.ndarray
        image or list of images
    cmap: string
        Optionnal, name of the colorbar to use.
    legend: string, or list of string
        legend under each image
    colorbar: bool
        Whether to show colorbar (default is True)
    share: bool
        if True, the axis are shared between the images, so that zooming in one image
        will zoom in all the corresponding regions of the other images. Default is True
    """
    if not (isinstance(img, (list, tuple))):
        img = [img]
    if isinstance(img, np.ndarray) and img.ndims == 3:
        img = [im for im in img]
    nimg = len(img)
    if legend is not None and not (isinstance(legend, (list, tuple))):
        legend = [legend]

    def _get_num_rows_cols(nimg):
        ncols = ceil(sqrt(nimg))
        nrows = ceil(nimg / ncols)
        return nrows, ncols

    subplot_kwargs = {}
    imshow_kwargs = {"interpolation": "nearest", "cmap": cmap}

    nrows, ncols = _get_num_rows_cols(nimg)
    plt.figure()
    for i in range(nimg):
        current_subplot = [nrows, ncols, i + 1]
        image = img[i]
        if rescale_n_std is not None:
            image = clip_std(image, n=rescale_n_std)
        if share and nimg > 1 and i == 0:
            ax0 = plt.subplot(*current_subplot)
            subplot_kwargs.update({"sharex": ax0, "sharey": ax0})
        else:
            plt.subplot(*current_subplot, **subplot_kwargs)

        plt.imshow(image, **imshow_kwargs)
        if legend:
            plt.xlabel(legend[i])
        if colorbar:
            plt.colorbar()

    plt.show()



import logging
import pytest
import numpy as np
from scipy.ndimage import gaussian_filter, shift
from nabu.utils import subdivide_into_overlapping_segment, clip_circle
from nabu.testutils import __do_long_tests__, generate_tests_scenarios

try:
    import astra

    __has_astra__ = True
except ImportError:
    __has_astra__ = False
from nabu.cuda.utils import __has_cupy__

if __has_cupy__:
    from nabu.reconstruction.cone import ConebeamReconstructor, NumpyConebeamReconstructor
if __has_astra__:
    from astra.extrautils import clipCircle


@pytest.fixture(scope="class")
def bootstrap(request):
    cls = request.cls
    cls.vol_shape = (128, 126, 126)
    cls.n_angles = 180
    cls.prj_width = 192  # detector larger than the sample
    cls.src_orig_dist = 1000
    cls.orig_det_dist = 100
    cls.volume, cls.cone_data = generate_hollow_cube_cone_sinograms(
        cls.vol_shape, cls.n_angles, cls.src_orig_dist, cls.orig_det_dist, prj_width=cls.prj_width
    )
    if __has_cupy__:
        ...


@pytest.mark.skipif(not (__has_cupy__ and __has_astra__), reason="Need cupy and astra for this test")
@pytest.mark.usefixtures("bootstrap")
class TestCone:
    def _create_cone_reconstructor(self, relative_z_position=None):
        return ConebeamReconstructor(
            self.cone_data.shape,
            self.src_orig_dist,
            self.orig_det_dist,
            relative_z_position=relative_z_position,
            volume_shape=self.volume.shape,
            # cuda_options={"ctx": self.ctx},
        )

    def test_simple_cone_reconstruction(self):
        C = self._create_cone_reconstructor()
        res = C.reconstruct(self.cone_data)
        delta = np.abs(res - self.volume)

        # Can we do better ? We already had to lowpass-filter the volume!
        # First/last slices are OK
        assert np.max(delta[:8]) < 1e-5
        assert np.max(delta[-8:]) < 1e-5
        # Middle region has a relatively low error
        assert np.max(delta[40:-40]) < 0.11
        # Transition zones between "zero" and "cube" has a large error
        assert np.max(delta[10:25]) < 0.2
        assert np.max(delta[-25:-10]) < 0.2
        # End of transition zones have a smaller error
        assert np.max(delta[25:40]) < 0.125
        assert np.max(delta[-40:-25]) < 0.125

    def test_against_explicit_astra_calls(self):
        C = self._create_cone_reconstructor()

        res = C.reconstruct(self.cone_data)
        #
        # Check that ConebeamReconstructor is consistent with these calls to astra
        #
        # "vol_geom" shape layout is (y, x, z). But here this geometry is used for the reconstruction
        # (i.e sinogram -> volume)and not for projection (volume -> sinograms).
        # So we assume a square slice. Mind that this is a particular case.
        vol_geom = astra.create_vol_geom(self.vol_shape[2], self.vol_shape[2], self.vol_shape[0])

        angles = np.linspace(0, 2 * np.pi, self.n_angles, True)
        proj_geom = astra.create_proj_geom(
            "cone",
            1.0,
            1.0,
            self.cone_data.shape[0],
            self.prj_width,
            angles,
            self.src_orig_dist,
            self.orig_det_dist,
        )
        sino_id = astra.data3d.create("-sino", proj_geom, data=self.cone_data)
        rec_id = astra.data3d.create("-vol", vol_geom)

        cfg = astra.astra_dict("FDK_CUDA")
        cfg["ReconstructionDataId"] = rec_id
        cfg["ProjectionDataId"] = sino_id
        alg_id = astra.algorithm.create(cfg)
        astra.algorithm.run(alg_id)

        res_astra = astra.data3d.get(rec_id)

        # housekeeping
        astra.algorithm.delete(alg_id)
        astra.data3d.delete(rec_id)
        astra.data3d.delete(sino_id)

        assert (
            np.max(np.abs(res - res_astra)) < 5e-4
        ), "ConebeamReconstructor results are inconsistent with plain calls to astra"

    def test_projection_full_vs_partial(self):
        """
        In the ideal case, all the data volume (and reconstruction) fits in memory.
        In practice this is rarely the case, so we have to reconstruct the volume slabs by slabs.
        The slabs should be slightly overlapping to avoid "stitching" artefacts at the edges.
        """
        # Astra seems to duplicate the projection data, even if all GPU memory is handled externally
        # Let's try with (n_z * n_y * n_x + 2 * n_a * n_z * n_x) * 4  <  mem_limit
        # 256^3 seems OK with n_a = 200 (180 MB)
        n_z = n_y = n_x = 256
        n_a = 200
        src_orig_dist = 1000
        orig_det_dist = 100

        volume, cone_data = generate_hollow_cube_cone_sinograms(
            vol_shape=(n_z, n_y, n_x), n_angles=n_a, src_orig_dist=src_orig_dist, orig_det_dist=orig_det_dist
        )
        C_full = ConebeamReconstructor(
            cone_data.shape, src_orig_dist, orig_det_dist
        )  # , cuda_options={"ctx": self.ctx})

        vol_geom = astra.create_vol_geom(n_y, n_x, n_z)

        proj_geom = astra.create_proj_geom("cone", 1.0, 1.0, n_z, n_x, C_full.angles, src_orig_dist, orig_det_dist)
        proj_id, projs_full_geom = astra.create_sino3d_gpu(volume, proj_geom, vol_geom)
        astra.data3d.delete(proj_id)

        # Do the same slab-by-slab
        inner_slab_size = 64
        overlap = 16
        slab_size = inner_slab_size + overlap * 2
        slabs = subdivide_into_overlapping_segment(n_z, slab_size, overlap)

        projs_partial_geom = np.zeros_like(projs_full_geom)
        for slab in slabs:
            z_min, z_inner_min, z_inner_max, z_max = slab
            rel_z_pos = (z_min + z_max) / 2 - n_z / 2
            subvolume = volume[z_min:z_max, :, :]
            C = ConebeamReconstructor(
                (z_max - z_min, n_a, n_x),
                src_orig_dist,
                orig_det_dist,
                relative_z_position=rel_z_pos,
                # cuda_options={"ctx": self.ctx},
            )
            proj_id, projs = astra.create_sino3d_gpu(subvolume, C.proj_geom, C.vol_geom)
            astra.data3d.delete(proj_id)

            projs_partial_geom[z_inner_min:z_inner_max] = projs[z_inner_min - z_min : z_inner_max - z_min]

        error_profile = [
            np.max(np.abs(proj_partial - proj_full))
            for proj_partial, proj_full in zip(projs_partial_geom, projs_full_geom)
        ]
        assert np.all(np.isclose(error_profile, 0.0, atol=0.0375)), "Mismatch between full-cone and slab geometries"

    def test_cone_reconstruction_magnified_vs_demagnified(self):
        """
        This will only test the astra toolbox.
        When reconstructing a volume from cone-beam data, the volume "should" have a smaller shape than the projection
        data shape (because of cone magnification).
        But astra provides the same results when backprojecting on a "de-magnified grid" and the original grid shape.
        """
        n_z = n_y = n_x = 256
        n_a = 500
        src_orig_dist = 1000
        orig_det_dist = 100
        magnification = 1 + orig_det_dist / src_orig_dist
        angles = np.linspace(0, 2 * np.pi, n_a, True)

        volume, cone_data = generate_hollow_cube_cone_sinograms(
            vol_shape=(n_z, n_y, n_x),
            n_angles=n_a,
            src_orig_dist=src_orig_dist,
            orig_det_dist=orig_det_dist,
            apply_filter=False,
        )
        rec_original_grid = astra_cone_beam_reconstruction(
            cone_data, angles, src_orig_dist, orig_det_dist, demagnify_volume=False
        )
        rec_reduced_grid = astra_cone_beam_reconstruction(
            cone_data, angles, src_orig_dist, orig_det_dist, demagnify_volume=True
        )

        m_z = (n_z - int(n_z / magnification)) // 2
        m_y = (n_y - int(n_y / magnification)) // 2
        m_x = (n_x - int(n_x / magnification)) // 2

        assert np.allclose(rec_original_grid[m_z:-m_z, m_y:-m_y, m_x:-m_x], rec_reduced_grid)

    def test_reconstruction_full_vs_partial(self):
        n_z = n_y = n_x = 256
        n_a = 500
        src_orig_dist = 1000
        orig_det_dist = 100
        angles = np.linspace(0, 2 * np.pi, n_a, True)

        volume, cone_data = generate_hollow_cube_cone_sinograms(
            vol_shape=(n_z, n_y, n_x),
            n_angles=n_a,
            src_orig_dist=src_orig_dist,
            orig_det_dist=orig_det_dist,
            apply_filter=False,
        )

        rec_full_volume = astra_cone_beam_reconstruction(cone_data, angles, src_orig_dist, orig_det_dist)

        rec_partial = np.zeros_like(rec_full_volume)
        inner_slab_size = 64
        overlap = 18
        slab_size = inner_slab_size + overlap * 2
        slabs = subdivide_into_overlapping_segment(n_z, slab_size, overlap)
        for slab in slabs:
            z_min, z_inner_min, z_inner_max, z_max = slab
            m1, m2 = z_inner_min - z_min, z_max - z_inner_max
            C = ConebeamReconstructor((z_max - z_min, n_a, n_x), src_orig_dist, orig_det_dist)
            rec = C.reconstruct(
                cone_data[z_min:z_max],
                relative_z_position=((z_min + z_max) / 2) - n_z / 2,  #  (z_min + z_max)/2.
            )
            rec_partial[z_inner_min:z_inner_max] = rec[m1 : (-m2) or None]

        # Compare volumes in inner circle
        for i in range(n_z):
            clipCircle(rec_partial[i])
            clipCircle(rec_full_volume[i])

        diff = np.abs(rec_partial - rec_full_volume)
        err_max_profile = np.max(diff, axis=(-1, -2))
        err_median_profile = np.median(diff, axis=(-1, -2))

        assert np.max(err_max_profile) < 2e-3
        assert np.max(err_median_profile) < 5.1e-6

    def test_reconstruction_horizontal_translations(self):
        n_z = n_y = n_x = 256
        n_a = 500
        src_orig_dist = 1000
        orig_det_dist = 50

        volume, cone_data = generate_hollow_cube_cone_sinograms(
            vol_shape=(n_z, n_y, n_x),
            n_angles=n_a,
            src_orig_dist=src_orig_dist,
            orig_det_dist=orig_det_dist,
            apply_filter=False,
        )

        # Apply horizontal translations on projections. This could have been done directly with astra
        shift_min, shift_max = -2, 5
        shifts_float = (shift_max - shift_min) * np.random.rand(n_a) - shift_min
        shifts_int = np.random.randint(shift_min, high=shift_max + 1, size=n_a)

        reconstructor_args = [
            cone_data.shape,
            src_orig_dist,
            orig_det_dist,
        ]
        reconstructor_kwargs = {
            "volume_shape": volume.shape,
            # "cuda_options": {"ctx": self.ctx},
        }
        cone_reconstructor = ConebeamReconstructor(*reconstructor_args, **reconstructor_kwargs)
        rec = cone_reconstructor.reconstruct(cone_data)

        # Translations done with floating-point shift values give a blurring of the image that cannot be recovered.
        # Error tolerance has to be higher for these shifts.
        for shift_type, shifts, err_tol in [
            ("integer shifts", shifts_int, 5e-3),
            ("float shifts", shifts_float, 1.6e-1),
        ]:
            cone_data_shifted = np.zeros_like(cone_data)
            [shift(cone_data[:, i, :], (0, shifts[i]), output=cone_data_shifted[:, i, :]) for i in range(n_a)]

            # Reconstruct with horizontal shifts
            cone_reconstructor_with_correction = ConebeamReconstructor(
                *reconstructor_args,
                **reconstructor_kwargs,
                extra_options={"axis_correction": -shifts},
            )

            rec_with_correction = cone_reconstructor_with_correction.reconstruct(cone_data_shifted)

            metric = lambda img: np.max(np.abs(clip_circle(img, radius=int(0.85 * img.shape[1] // 2))))
            error_profile = np.array([metric(rec[i] - rec_with_correction[i]) for i in range(n_z)])
            assert error_profile.max() < err_tol, "Max error with %s is too high" % shift_type

    def test_padding_mode(self):
        n_z = n_y = n_x = 256
        n_a = 500
        src_orig_dist = 1000
        orig_det_dist = 50

        volume, cone_data = generate_hollow_cube_cone_sinograms(
            vol_shape=(n_z, n_y, n_x),
            n_angles=n_a,
            src_orig_dist=src_orig_dist,
            orig_det_dist=orig_det_dist,
            apply_filter=False,
        )
        reconstructor_args = [
            cone_data.shape,
            src_orig_dist,
            orig_det_dist,
        ]
        reconstructor_kwargs = {
            "volume_shape": volume.shape,
            # "cuda_options": {"ctx": self.ctx},
        }
        cone_reconstructor_zero_padding = ConebeamReconstructor(*reconstructor_args, **reconstructor_kwargs)
        rec_z = cone_reconstructor_zero_padding.reconstruct(cone_data)

        for padding_mode in ["edges"]:
            cone_reconstructor = ConebeamReconstructor(
                *reconstructor_args, padding_mode=padding_mode, **reconstructor_kwargs
            )
            rec = cone_reconstructor.reconstruct(cone_data)

            metric = lambda img: np.max(np.abs(clip_circle(img, radius=int(0.85 * 128))))
            error_profile = np.array([metric(rec[i] - rec_z[i]) for i in range(n_z)])

            # import matplotlib.pyplot as plt
            # plt.figure()
            # plt.plot(np.arange(n_z), error_profile)
            # plt.legend([padding_mode])
            # plt.show()

            assert error_profile.max() < 3.1e-2, "Max error for padding=%s is too high" % padding_mode
            if padding_mode != "zeros":
                assert not (np.allclose(rec[n_z // 2], rec_z[n_z // 2])), (
                    "Reconstruction should be different when padding_mode=%s" % padding_mode
                )

    def test_roi(self):
        n_z = n_y = n_x = 256
        n_a = 500
        src_orig_dist = 1000
        orig_det_dist = 50

        volume, cone_data = generate_hollow_cube_cone_sinograms(
            vol_shape=(n_z, n_y, n_x),
            n_angles=n_a,
            src_orig_dist=src_orig_dist,
            orig_det_dist=orig_det_dist,
            apply_filter=False,
            rot_center_shift=10,
        )

        reconstructor_args = [
            cone_data.shape,
            src_orig_dist,
            orig_det_dist,
        ]
        reconstructor_kwargs = {
            "volume_shape": volume.shape,
            "rot_center": (n_x - 1) / 2 + 10,
            # "cuda_options": {"ctx": self.ctx},
        }
        cone_reconstructor_full = ConebeamReconstructor(*reconstructor_args, **reconstructor_kwargs)
        ref = cone_reconstructor_full.reconstruct(cone_data)

        # roi is in the form (start_x, end_x, start_y, end_y)
        for roi in ((20, -20, 10, -10), (0, n_x, 0, n_y), (50, -50, 15, -15)):
            # convert negative indices
            start_x, end_x, start_y, end_y = roi
            if start_y < 0:
                start_y += n_y
            if start_x < 0:
                start_x += n_x

            cone_reconstructor = ConebeamReconstructor(*reconstructor_args, slice_roi=roi, **reconstructor_kwargs)
            rec = cone_reconstructor.reconstruct(cone_data)

            assert np.allclose(rec, ref[:, roi[2] : roi[3], roi[0] : roi[1]]), "Something wrong with roi=%s" % (
                str(roi)
            )

    def test_fdk_preweight(self, caplog):
        """
        Check that nabu's FDK pre-weighting give the same results as astra
        """
        shapes = [
            {"n_z": 256, "n_x": 256, "n_a": 500},
            # {"n_z": 250, "n_x": 340, "n_a": 250}, # Astra reconstruction is incorrect in this case!
        ]
        src_orig_dist = 1000
        orig_det_dist = 50

        rot_centers_from_middle = [0]
        if __do_long_tests__:
            rot_centers_from_middle.extend([10, -15])

        params_list = generate_tests_scenarios({"shape": shapes, "rot_center": rot_centers_from_middle})

        for params in params_list:
            n_z = params["shape"]["n_z"]
            n_x = n_y = params["shape"]["n_x"]
            n_a = params["shape"]["n_a"]
            rc = params["rot_center"]
            volume, cone_data = generate_hollow_cube_cone_sinograms(
                vol_shape=(n_z, n_y, n_x),
                n_angles=n_a,
                src_orig_dist=src_orig_dist,
                orig_det_dist=orig_det_dist,
                apply_filter=False,
                rot_center_shift=rc,
            )

            reconstructor_args = [(n_z, n_a, n_x), src_orig_dist, orig_det_dist]
            reconstructor_kwargs_base = {
                "volume_shape": volume.shape,
                "rot_center": (n_x - 1) / 2 + rc,
                # "cuda_options": {"ctx": self.ctx},
            }
            reconstructor_kwargs_astra = {"padding_mode": "zeros", "extra_options": {"use_astra_fdk": True}}
            reconstructor_kwargs_nabu = {"padding_mode": "zeros", "extra_options": {"use_astra_fdk": False}}
            reconstructor_astra = ConebeamReconstructor(
                *reconstructor_args, **{**reconstructor_kwargs_base, **reconstructor_kwargs_astra}
            )
            assert reconstructor_astra._use_astra_fdk is True, "reconstructor_astra should use native astra FDK"
            reconstructor_nabu = ConebeamReconstructor(
                *reconstructor_args, **{**reconstructor_kwargs_base, **reconstructor_kwargs_nabu}
            )
            ref = reconstructor_astra.reconstruct(cone_data)
            res = reconstructor_nabu.reconstruct(cone_data)

            reconstructor_kwargs_nabu = {"padding_mode": "edges", "extra_options": {"use_astra_fdk": False}}
            cb_ep = ConebeamReconstructor(
                *reconstructor_args, **{**reconstructor_kwargs_base, **reconstructor_kwargs_nabu}
            )
            res_ep = cb_ep.reconstruct(cone_data)  # noqa: F841

            assert np.max(np.abs(res - ref)) < 2e-3, "Wrong FDK results for parameters: %s" % (str(params))

            # Test with edges padding - only nabu can do that
            reconstructor_kwargs_nabu["padding_mode"] = "edges"
            reconstructor_nabu = ConebeamReconstructor(
                *reconstructor_args, **{**reconstructor_kwargs_base, **reconstructor_kwargs_nabu}
            )
            reconstructor_nabu.reconstruct(cone_data)
            # result is slightly different than "res" in the borders, which is expected
            # it would be good to test it as well, but it's outside of the scope of this test

            with caplog.at_level(logging.WARNING):
                reconstructor_kwargs_nabu = {"padding_mode": "edges", "extra_options": {"use_astra_fdk": True}}
                ConebeamReconstructor(*reconstructor_args, **{**reconstructor_kwargs_base, **reconstructor_kwargs_nabu})
                assert "cannot use native astra FDK" in caplog.text

    def test_reconstruct_noncontiguous_data(self):
        n_z = 206
        n_y = n_x = 256
        n_a = 500
        src_orig_dist = 1000
        orig_det_dist = 50

        volume, cone_data = generate_hollow_cube_cone_sinograms(
            vol_shape=(n_z, n_y, n_x),
            n_angles=n_a,
            src_orig_dist=src_orig_dist,
            orig_det_dist=orig_det_dist,
            apply_filter=False,
            rot_center_shift=10,
        )
        cone_reconstructor = ConebeamReconstructor(
            cone_data.shape,
            src_orig_dist,
            orig_det_dist,
            volume_shape=volume.shape,
            rot_center=(n_x - 1) / 2 + 10,
            # cuda_options={"ctx": self.ctx},
            extra_options={"use_astra_fdk": False},
        )
        ref = cone_reconstructor.reconstruct(cone_data)

        radios = cone_reconstructor.cuda.allocate_array("_radios", (n_a, n_z, n_x))
        for i in range(n_a):
            radios[i].set(cone_data[:, i, :])

        sinos_discontig = radios.transpose((1, 0, 2))
        assert cone_reconstructor.cuda.is_contiguous(sinos_discontig) is False
        res = cone_reconstructor.reconstruct(sinos_discontig)
        assert np.allclose(res, ref), "Reconstructing non-contiguous data failed"

    def test_cone_clip_circle(self):
        n_z = 206
        n_y = n_x = 256
        n_a = 500
        src_orig_dist = 1000
        orig_det_dist = 50

        volume, cone_data = generate_hollow_cube_cone_sinograms(
            vol_shape=(n_z, n_y, n_x),
            n_angles=n_a,
            src_orig_dist=src_orig_dist,
            orig_det_dist=orig_det_dist,
            apply_filter=False,
            rot_center_shift=10,
        )
        init_args = [cone_data.shape, src_orig_dist, orig_det_dist]
        init_kwargs = {
            "volume_shape": volume.shape,
            "rot_center": (n_x - 1) / 2 + 10,
            "extra_options": {"use_astra_fdk": False, "clip_outer_circle": True},
        }

        # for cone_reconstructor_cls in [ConebeamReconstructor, NumpyConebeamReconstructor]:
        for cone_reconstructor_cls in [NumpyConebeamReconstructor]:
            for out_circle_value in [0, np.nan]:
                init_kwargs["extra_options"]["outer_circle_value"] = out_circle_value
                cone_reconstructor = cone_reconstructor_cls(*init_args, **init_kwargs)
                rec = cone_reconstructor.reconstruct(cone_data)
                for i in range(rec.shape[0]):
                    err_msg = f"Something wrong with clip_circle for {cone_reconstructor.__class__.__name__} with outer value {out_circle_value} at slice number {i}"
                    assert np.allclose(clip_circle(rec[i], out_value=out_circle_value), rec[i], equal_nan=True), err_msg

    def _test_numpy_cone(self, cor_shift, padding_mode):
        d1 = 1000
        d2 = 10

        rot_center = None
        neg_cor_shift = None
        cube, sinos = generate_hollow_cube_cone_sinograms((256, 256, 200), 200, d1, d2, rot_center_shift=cor_shift)

        if cor_shift is not None:
            rot_center = (sinos.shape[-1] - 1) / 2 + cor_shift
            neg_cor_shift = -cor_shift
        rec0 = astra_cone_beam_reconstruction(
            sinos, np.linspace(0, 2 * np.pi, sinos.shape[1], True), d1, d2, cor_shift=neg_cor_shift
        )

        cone_reconstructor = NumpyConebeamReconstructor(
            sinos.shape, d1, d2, rot_center=rot_center, padding_mode=padding_mode
        )
        rec = cone_reconstructor.reconstruct(sinos)

        err_profile = np.max(np.abs(rec - rec0), axis=(1, 2))
        tol = 2e-3

        if padding_mode not in ["zeros", "constant"]:
            # Astra does not support natively padding mode other than zeros
            # Here we ignore the region close to the edges, since edges-padding gives a different result
            err_profile = np.max(np.abs(rec[:, 50:-50, 50:-50] - rec0[:, 50:-50, 50:-50]), axis=(1, 2))[50:-50]
            tol = 0.15

        assert (
            np.max(err_profile) < tol
        ), f"NumpyConebeamReconstructor: FDK reconstruction failed for cor_shift={cor_shift}, padding_mode={padding_mode}"

    def test_numpy_cone(self):
        test_cases = {
            "cor_shift": [None, 6.5],
            "padding_mode": ["zeros", "edges"],
        }
        params_list = generate_tests_scenarios(test_cases)

        for params in params_list:
            self._test_numpy_cone(params["cor_shift"], params["padding_mode"])


def generate_hollow_cube_cone_sinograms(
    vol_shape,
    n_angles,
    src_orig_dist,
    orig_det_dist,
    prj_width=None,
    apply_filter=True,
    rot_center_shift=None,
):
    # Adapted from Astra toolbox python samples

    n_z, n_y, n_x = vol_shape
    vol_geom = astra.create_vol_geom(n_y, n_x, n_z)

    prj_width = prj_width or n_x
    # prj_height = n_z
    angles = np.linspace(0, 2 * np.pi, n_angles, True)

    proj_geom = astra.create_proj_geom("cone", 1.0, 1.0, n_z, prj_width, angles, src_orig_dist, orig_det_dist)
    if rot_center_shift is not None:
        proj_geom = astra.geom_postalignment(proj_geom, (-rot_center_shift, 0))
    # magnification = 1 + orig_det_dist / src_orig_dist

    # hollow cube
    cube = np.zeros(astra.geom_size(vol_geom), dtype="f")

    d = int(min(n_x, n_y) / 2 * (1 - np.sqrt(2) / 2))
    cube[20:-20, d:-d, d:-d] = 1
    cube[40:-40, d + 20 : -(d + 20), d + 20 : -(d + 20)] = 0

    # d = int(min(n_x, n_y) / 2 * (1 - np.sqrt(2) / 2) * magnification)
    # d1 = d + 10
    # d2 = d + 20
    # cube[40:-40, d1:-d1, d1:-d1] = 1
    # cube[60:-60, d2 : -d2, d2 : -d2] = 0

    # High-frequencies yield cannot be accurately retrieved
    if apply_filter:
        cube = gaussian_filter(cube, (1.0, 1.0, 1.0))

    proj_id, proj_data = astra.create_sino3d_gpu(cube, proj_geom, vol_geom)
    astra.data3d.delete(proj_id)  # (n_z, n_angles, n_x)

    return cube, proj_data


def astra_cone_beam_reconstruction(
    cone_data, angles, src_orig_dist, orig_det_dist, demagnify_volume=False, cor_shift=None
):
    """
    Handy (but data-inefficient) function to reconstruct data from cone-beam geometry
    """

    n_z, n_a, n_x = cone_data.shape

    proj_geom = astra.create_proj_geom("cone", 1.0, 1.0, n_z, n_x, angles, src_orig_dist, orig_det_dist)
    if cor_shift is not None:
        proj_geom = astra.geom_postalignment(proj_geom, (cor_shift, 0))
    sino_id = astra.data3d.create("-sino", proj_geom, data=cone_data)

    m = 1 + orig_det_dist / src_orig_dist if demagnify_volume else 1.0
    n_z_vol, n_y_vol, n_x_vol = int(n_z / m), int(n_x / m), int(n_x / m)
    vol_geom = astra.create_vol_geom(n_y_vol, n_x_vol, n_z_vol)
    rec_id = astra.data3d.create("-vol", vol_geom)

    cfg = astra.astra_dict("FDK_CUDA")
    cfg["ReconstructionDataId"] = rec_id
    cfg["ProjectionDataId"] = sino_id
    alg_id = astra.algorithm.create(cfg)

    astra.algorithm.run(alg_id)

    rec = astra.data3d.get(rec_id)

    astra.data3d.delete(sino_id)
    astra.data3d.delete(rec_id)
    astra.algorithm.delete(alg_id)

    return rec



from pathlib import Path
from os import path, mkdir, rename
import numpy as np
import pytest
from tomoscan.esrf.volume import (
    EDFVolume,
    HDF5Volume,
    JP2KVolume,
    MultiTIFFVolume,
    TIFFVolume,
)
from tomoscan.esrf.volume.jp2kvolume import has_glymur
from nabu.io.writer import merge_hdf5_files
from nabu.io.cast_volume import remove_volume


def test_remove_single_frame_volume(tmpdir):
    """
    Test volume removal for tiff, jp2 and EDF
    """
    # Have to use a not-too-small size because of jp2k
    data = np.arange(10 * 40 * 50, dtype="f").reshape((10, 40, 50))

    volume_classes = [EDFVolume, TIFFVolume, JP2KVolume]
    if not (has_glymur):
        volume_classes.pop()
    for volume_cls in volume_classes:
        ext = volume_cls.DEFAULT_DATA_EXTENSION
        folder = path.join(tmpdir, f"{ext}_vol")
        volume_basename = f"{ext}_basename"

        vol_writer = volume_cls(folder=folder, volume_basename=volume_basename, overwrite=True, start_index=0)
        vol_writer.data = data
        vol_writer.save()

        vol_reader = volume_cls(folder=folder, volume_basename=volume_basename)
        assert path.isdir(folder), f"Expected to find a folder f{folder}"
        remove_volume(vol_reader)
        assert not (path.isdir(folder)), f"Expected to have removed the folder f{folder}"

        vol_writer.save()
        vol_reader = volume_cls(folder=folder, volume_basename=volume_basename)
        Path(path.join(folder, f"unexpected.{ext}")).touch()
        with pytest.raises(RuntimeError) as exc:
            remove_volume(vol_reader, check=True)
        assert "Unexpected files present" in str(exc.value), "Expected check to find extraneous files"


def test_remove_multiframe_volume(tmpdir):
    """
    Test volume removal for "multiframe" formats (HDF5, tiff3D)
    The HDF5 files considered in this test do not have virtual sources
    """
    data = np.arange(3 * 4 * 5, dtype="f").reshape((3, 4, 5))

    for ext, volume_cls in {"h5": HDF5Volume, "tiff": MultiTIFFVolume}.items():
        file_path = path.join(tmpdir, f"{ext}_vol.{ext}")

        init_kwargs = {"file_path": file_path}
        if ext == "h5":
            init_kwargs["data_path"] = "entry"
        vol_writer = volume_cls(**init_kwargs)
        vol_writer.data = data
        vol_writer.save()

        vol_reader = volume_cls(**init_kwargs)
        assert path.isfile(file_path), f"Expected to find a {ext} volume at {file_path}"
        remove_volume(vol_reader)
        assert not (path.isfile(file_path)), f"Expected to have removed f{file_path}"


def test_remove_hdf5_multiple_entries(tmpdir):
    data = np.arange(3 * 4 * 5, dtype="f").reshape((3, 4, 5))
    file_path = path.join(tmpdir, "h5_vol.h5")
    vol_writer_1 = HDF5Volume(file_path=file_path, data_path="entry0000")
    vol_writer_1.data = data
    vol_writer_1.save()
    vol_writer_2 = HDF5Volume(file_path=file_path, data_path="entry0001")
    vol_writer_2.data = data + 10
    vol_writer_2.save()
    vol_reader = HDF5Volume(file_path=file_path, data_path="entry0000")
    with pytest.raises(NotImplementedError) as exc:
        remove_volume(vol_reader, check=True)
    assert "Removing a HDF5 volume with more than one entry is not supported" in str(
        exc.value
    ), "Expected an error message"


def test_remove_nabu_hdf5_reconstruction(tmpdir):
    """
    Test removal of HDF5 reconstruction generated by nabu (i.e with virtual sources)
    """

    entry = "entry"
    process_name = "reconstruction"

    master_file_path = path.join(tmpdir, "sample_naburec.hdf5")
    associated_dir = path.join(tmpdir, "sample_naburec")
    if not (path.isdir(associated_dir)):
        mkdir(associated_dir)

    n_chunks = 5
    local_files = []
    for i in range(n_chunks):
        fname = "sample_naburec_%06d.h5" % i
        partial_rec_abspath = path.join(associated_dir, fname)
        local_files.append(f"sample_naburec/{fname}")
        # local_files.append(fname)
        vol = HDF5Volume(partial_rec_abspath, data_path=f"{entry}/{process_name}")
        vol.data = np.arange(3 * 4 * 5, dtype="f").reshape((3, 4, 5))
        vol.save()

    h5_path = f"{entry}/{process_name}/results/data"

    merge_hdf5_files(
        local_files,
        h5_path,
        master_file_path,
        process_name,
        output_entry=entry,
        output_filemode="a",
        processing_index=0,
        config=None,
        base_dir=path.dirname(associated_dir),
        axis=0,
        overwrite=True,
    )

    assert path.isfile(master_file_path), f"Expected to find the master file at {master_file_path}"
    assert path.isdir(associated_dir)
    for local_file in local_files:
        partial_rec_file = path.join(tmpdir, local_file)
        assert path.isfile(partial_rec_file), f"Expected to find partial file number {i} at {partial_rec_file}"

    # Check that the virtual links are handled properly
    #  sample_rec.hdf5   should reference sample_rec/sample_rec_{i}.h5
    renamed_master_file_path = (
        path.join(path.dirname(master_file_path), path.basename(master_file_path).split(".")[0]) + "_renamed" + ".h5"
    )
    rename(master_file_path, renamed_master_file_path)
    h5_vol = HDF5Volume(file_path=renamed_master_file_path, data_path=f"{entry}/{process_name}")
    with pytest.raises(ValueError) as exc:
        remove_volume(h5_vol)
    expected_error_message = f"The virtual sources in {renamed_master_file_path}:{process_name}/results/data reference the directory sample_naburec, but expected was sample_naburec_renamed"
    assert str(exc.value) == expected_error_message

    # Check removal in normal circumstances
    rename(renamed_master_file_path, master_file_path)
    h5_vol = HDF5Volume(file_path=master_file_path, data_path=f"{entry}/{process_name}")
    remove_volume(h5_vol)
    assert not (path.isfile(master_file_path)), f"Expected to find the master file at {master_file_path}"
    assert not (path.isdir(associated_dir))



import os
import warnings
from importlib import import_module
from functools import lru_cache
from multiprocessing import get_context
from multiprocessing.pool import Pool
from ..utils import BaseClassError, check_supported, no_decorator
from .fft_base import _BaseFFT, _BaseVKFFT
from ..cuda.processing import CudaProcessing

try:
    from pyvkfft.cuda import VkFFTApp as CudaVkFFTApp

    __has_vkfft__ = True
except (ImportError, OSError):
    __has_vkfft__ = False
    CudaVkFFTApp = BaseClassError

try:
    import cupy.fft as cupy_fft

    __has_cupy__ = True
except (ImportError, OSError):
    __has_cupy__ = False


n_cached_ffts = int(os.getenv("NABU_FFT_CACHE", "0"))


maybe_cached = lru_cache(maxsize=n_cached_ffts) if n_cached_ffts > 0 else no_decorator


@maybe_cached
def _get_vkfft_cuda(*args, **kwargs):
    return CudaVkFFTApp(*args, **kwargs)


def get_vkfft_cuda(slf, *args, **kwargs):
    return _get_vkfft_cuda(*args, **kwargs)


class VKCUFFT(_BaseVKFFT):
    """
    Cuda FFT, using VKFFT backend
    """

    implem = "vkfft"
    backend = "cuda"
    ProcessingCls = CudaProcessing
    get_fft_obj = get_vkfft_cuda

    def _init_backend(self, backend_options):
        super()._init_backend(backend_options)
        self._vkfft_other_init_kwargs = {"stream": self.processing.stream}


def _has_fft_implem(class_name, module_name, component_check=None):
    # should be run from within a Process
    try:
        mod = import_module(module_name)  # eg. 'import nabu.processing.fft_cuda'
        if component_check is not None and getattr(mod, component_check, False) is False:  # eg. '__has_vkfft__'
            return False
        fft_class = getattr(mod, class_name)

        from nabu.cuda.utils import get_cuda_stream

        _ = get_cuda_stream(force_create=True)
        _ = fft_class((16,), "f")
        avail = True
    except (ImportError, RuntimeError, OSError, NameError) as exc:
        avail = False
        print(exc)
    return avail


def has_fft_implem(name, safe=True):
    if not safe:
        return _has_fft_implem(name, "nabu.processing.fft_cuda")
    try:
        ctx = get_context("spawn")
        with Pool(1, context=ctx) as p:
            v = p.apply(_has_fft_implem, args=(name, "nabu.processing.fft_cuda"))
    except AssertionError:
        # Can get AssertionError: daemonic processes are not allowed to have children
        # if the calling code is already a subprocess
        return _has_fft_implem(name, "nabu.processing.fft_cuda")
    return v


@lru_cache(maxsize=2)
def has_vkfft(safe=True):
    return has_fft_implem("VKCUFFT", safe=safe)


@lru_cache(maxsize=2)
def has_cupyfft(safe=True):
    return has_fft_implem("CupyCUFFT", safe=safe)


@lru_cache(maxsize=2)
def get_fft_class(backend="vkfft"):
    backends = {
        "vkfft": VKCUFFT,
        "pyvkfft": VKCUFFT,
        "cupy": CupyCUFFT,
        "cufft": CupyCUFFT,
    }

    def get_fft_cls(asked_fft_backend):
        asked_fft_backend = asked_fft_backend.lower()
        check_supported(asked_fft_backend, list(backends.keys()), "Cuda FFT backend name")
        return backends[asked_fft_backend]

    asked_fft_backend_env = os.environ.get("NABU_FFT_BACKEND", "")
    if asked_fft_backend_env != "":
        return get_fft_cls(asked_fft_backend_env)

    avail_fft_implems = get_available_fft_implems()
    if len(avail_fft_implems) == 0:
        raise RuntimeError("Could not any Cuda FFT implementation. Please install pyvkfft and/or cupy")
    if backend not in avail_fft_implems:
        warnings.warn("Could not get FFT backend '%s'" % backend, RuntimeWarning)
        backend = avail_fft_implems[0]

    return get_fft_cls(backend)


@lru_cache(maxsize=1)
def get_available_fft_implems():
    avail_implems = []
    if has_vkfft(safe=True):
        avail_implems.append("vkfft")
    if has_cupyfft(safe=True):
        avail_implems.append("cupy")
    return avail_implems


class CupyCUFFT(_BaseFFT):

    implem = "cupy"
    backend = "cuda"
    ProcessingCls = CudaProcessing

    def _configure_batched_transform(self):
        # See documentation. Use this if we want to lessen the memory footprint at the expense of performances
        # cupy.fft.config.enable_nd_planning = False
        ...

    def _patch_cupy_ifft_kwargs(self, ifft_kwargs):
        """
        Patch cupy.irfft for odd-sized last dimensions.

        cupy irfft does not yield the same array shape as numpy/vkfft
          eg:   cupy.fft.irfft2(cp.fft(dz)).shape != np.fft.ifft2(np.fft.fft2(hz)
        numpy/vkfft are able to fall-back on the original array shape, but for cupy,
        we have to use the "s=" argument to fall back on the original data shape.
        """
        fft_ndim = len(self.axes)
        kwarg_key = "n" if fft_ndim == 1 else "s"

        transf_axis_shp = self.get_transformed_axes_shape()
        if len(transf_axis_shp) == 1:
            transf_axis_shp = transf_axis_shp[0]
        ifft_kwargs[kwarg_key] = transf_axis_shp

    def _compute_fft_plans(self):
        data_ndim = len(self.shape)
        fft_ndim = len(self.axes)
        fft_kwargs = {}
        if data_ndim > 1:
            if fft_ndim == 1:
                fft_kwargs = {"axis": self.axes[0]}
            else:
                fft_kwargs = {"axes": self.axes}
        ifft_kwargs = fft_kwargs.copy()
        self._patch_cupy_ifft_kwargs(ifft_kwargs)

        self._fft_kwargs = fft_kwargs
        self._ifft_kwargs = ifft_kwargs

        if fft_ndim == 1:
            fft_func = cupy_fft.fft if not (self.r2c) else cupy_fft.rfft
            ifft_func = cupy_fft.ifft if not (self.r2c) else cupy_fft.irfft
        elif fft_ndim == 2:
            fft_func = cupy_fft.fft2 if not (self.r2c) else cupy_fft.rfft2
            ifft_func = cupy_fft.ifft2 if not (self.r2c) else cupy_fft.irfft2
        else:
            fft_func = cupy_fft.fftn if not (self.r2c) else cupy_fft.rfftn
            ifft_func = cupy_fft.ifftn if not (self.r2c) else cupy_fft.irfftn
        self._fft_func = fft_func
        self._ifft_func = ifft_func

    def fft(self, array, output=None):
        out = self._fft_func(array, **self._fft_kwargs)
        if output is not None:
            output[:] = out[:]
        return output if output is not None else out

    def ifft(self, array, output=None):
        print(self._ifft_kwargs)
        out = self._ifft_func(array, **self._ifft_kwargs)
        if output is not None:
            output[:] = out[:]
        return output if output is not None else out



from warnings import warn
import numpy as np

from ..utils import check_supported
from ..resources.gpu import GPUDescription

try:
    from cupy import cuda, ndarray as cupy_ndarray

    __has_cupy__ = cuda.is_available()
    __cupy_error_msg__ = None
except ImportError as err:
    __has_cupy__ = False
    __cupy_error_msg__ = str(err)


# -----------------------------------------------------------------------------
# -------------------- Device management --------------------------------------
# -----------------------------------------------------------------------------


# TODO use existing context or stream
# cupy does not seem to be able to access contexts
def get_cuda_stream(device_id=None, force_create=False):
    """
    Notes
    ------
    In cupy, contexts/device/stream management is not so explicit.
    It seems that cupy wants to hide these things from user and handle it automatically.
    There are multiple issues:
      - cuda.get_current_stream() will always return device_id = -1
        (even when called with device_id != -1)
      - A Device object does not have a list of attached streams
      - A Stream object does not have its device, we have keep various objects references

    Solution so far:
      - Call Device(id).use() at program startup
      - Use cuda.runtime.getDevice()
      - Use cuda.stream.Stream()

    """
    if device_id is not None:
        dev = cuda.Device(device_id)
        dev.use()
    if force_create:
        return cuda.Stream(null=False, ptds=False)
    return cuda.Stream(null=False, ptds=True)


def detect_cuda_gpus():
    """
    Detect the available Nvidia CUDA GPUs on the current host.

    Returns
    --------
    gpus: dict
        Dictionary where the key is the GPU ID, and the value is a `cupy.cuda.Device` object.
    error_msg: str
        In the case where there is an error, the message is returned in this item.
        Otherwise, it is a None object.
    """
    gpus = {}
    error_msg = None
    if not (__has_cupy__):
        return {}, __cupy_error_msg__
    try:
        n_gpus = cuda.runtime.getDeviceCount()
    except Exception as exc:
        error_msg = str(exc)
    if error_msg is not None:
        return {}, error_msg
    for i in range(n_gpus):
        gpus[i] = cuda.Device(i)
    return gpus, None


def collect_cuda_gpus(on_error="warn"):
    """
    Return a dictionary of GPU ids and brief description of each CUDA-compatible
    GPU with a few fields.

    Parameters
    ----------
    on_error: str, optional
        What to do on error. Possible values:
          - "warn": print a warning
          - "raise": throw an error
    """
    gpus, error_msg = detect_cuda_gpus()
    if error_msg is not None:
        msg = "Failed to collect CUDA GPUs: %s" % error_msg
        if on_error == "warn":
            warn(msg, RuntimeWarning)
        elif on_error == "raise":
            raise RuntimeError(msg)
        return None
    cuda_gpus = {}
    for gpu_id, gpu in gpus.items():
        cuda_gpus[gpu_id] = GPUDescription(gpu).get_dict()
    return cuda_gpus


# -----------------------------------------------------------------------------
# -------------------- Data type utils  ---------------------------------------
# -----------------------------------------------------------------------------


_dtype_to_ctype_cuda = {
    np.bool_: "bool",
    np.byte: "signed char",  # same as int8
    np.ubyte: "unsigned char",  # same as uint8
    np.short: "short",  # same as int16
    np.ushort: "unsigned short",  # same as uint16
    np.intc: "int",  # typically int32
    np.uintc: "unsigned int",  # typically uint32
    np.int_: "long",  # platform-dependent, usually int64 on 64-bit Linux
    np.uint: "unsigned long",  # same logic as above
    np.longlong: "long long",  # explicitly int64
    np.ulonglong: "unsigned long long",  # explicitly uint64
    np.half: "half",  # not standard C, sometimes `_Float16` or compiler extension
    np.float16: "half",  # same as above
    np.single: "float",  # float32
    np.float32: "float",
    np.double: "double",  # float64
    np.float64: "double",
    np.longdouble: "long double",  # platform-dependent extended precision
    np.csingle: "complex64",  # "float _Complex",  # complex64
    np.complex64: "complex64",  # "float _Complex",
    np.cdouble: "complex128",  # "double _Complex",  # complex128
    np.complex128: "complex128",  # "double _Complex",
    np.clongdouble: "long double _Complex",  # complex256 on some platforms
}


def dtype_to_ctype(dtype):
    dtype = np.dtype(dtype).type  # Normalize
    return _dtype_to_ctype_cuda.get(dtype, None)


def to_int2(arr):
    """
    Convert a 1D array of length 2 into a "int2" data type.
    Beware, the first coordinate is x ! (as opposed to usual python/numpy/C convention)
    """
    res = np.zeros(1, dtype=int2_t)
    res["x"] = arr[0]
    res["y"] = arr[1]
    return res


# -----------------------------------------------------------------------------
# ---------------------- Array & Textures utils -------------------------------
# -----------------------------------------------------------------------------


def create_texture(shape, dtype, from_image=None, address_mode="border", filter_mode="linear", normalized_coords=False):
    """
    Create a texture object.

    Parameters
    ----------
    shape: tuple of int
        Image shape
    dtype: numpy.dtype
        Data type
    from_image: numpy.ndarray, optional
        Image to be used as a texture. If provided, above parameters are replaced with image properties.
    address_mode: str, optional
        Which address mode to use. Can be:
           - "border" (default): extension with zeros
           - "clamp": extension with edges
           - "mirror": extension with mirroring
           - "wrap": periodic extension
    filter_mode: str, optional
        Which filter mode to use when accessing texture at non-integer coordinates.
        Can be "linear" or "nearest"
        Default is (bi)linear filtering.
    normalized_coords: bool, optional
        Whether to use normalized coordinates. Default is False.

    Returns
    --------
    tex_obj: cupy.cuda.texture.TextureObject
        Texture object that can be passed to kernels (accessed in source code with 'cudaTextureObject_t' type)
    cu_arr: cupy.cuda.texture.CUDAarray
        Cuda array that bakes the texture

    Notes
    -----
      - Two-dimensional images default to single-chanel textures.
      - It will allocate memory for the underlying "cuda array" object
    """
    if from_image is not None:
        shape = from_image.shape
        dtype = from_image.dtype
    if len(shape) > 2:
        raise NotImplementedError("Textures > 2D are not supported yet")
    dtype_kind_to_format_kind = {
        "f": cuda.runtime.cudaChannelFormatKindFloat,
        "u": cuda.runtime.cudaChannelFormatKindUnsigned,
        "i": cuda.runtime.cudaChannelFormatKindSigned,
    }
    check_supported(np.dtype(dtype).kind, dtype_kind_to_format_kind, "data type")
    address_modes = {
        "border": cuda.runtime.cudaAddressModeBorder,  # mode="zeros"
        "clamp": cuda.runtime.cudaAddressModeClamp,  # mode="edges"
        "mirror": cuda.runtime.cudaAddressModeMirror,  # mirror
        "wrap": cuda.runtime.cudaAddressModeWrap,  # periodic
    }
    check_supported(address_mode, address_modes, "address mode")
    filter_modes = {
        "linear": cuda.runtime.cudaFilterModeLinear,
        "nearest": cuda.runtime.cudaFilterModePoint,
    }
    check_supported(filter_mode, filter_modes, "filter mode")

    tex_format = dtype_kind_to_format_kind[np.dtype(dtype).kind]
    nbits = np.dtype(dtype).itemsize * 8
    ch_desc = cuda.texture.ChannelFormatDescriptor(nbits, 0, 0, 0, tex_format)

    # /!\ From the documentation:
    # /!\ The memory allocation of CUDAarray is done outside of CuPy's memory management (enabled by default)
    # due to CUDA's limitation. Users of CUDAarray should be cautious about any out-of-memory possibilities.
    cu_arr = cuda.texture.CUDAarray(ch_desc, width=shape[1], height=shape[0], flags=cuda.runtime.cudaArrayDefault)

    if from_image is not None:
        # NB. For 2D textures, copy_from expects shape (height, nch*width); nch=1 here
        cu_arr.copy_from(from_image)

    res_desc = cuda.texture.ResourceDescriptor(cuda.runtime.cudaResourceTypeArray, cuArr=cu_arr)

    read_mode = cuda.runtime.cudaReadModeNormalizedFloat if normalized_coords else cuda.runtime.cudaReadModeElementType
    tex_desc = cuda.texture.TextureDescriptor(
        addressModes=(address_modes[address_mode], address_modes[address_mode]),
        filterMode=filter_modes[filter_mode],
        readMode=read_mode,
        normalizedCoords=normalized_coords,
    )

    tex_obj = cuda.texture.TextureObject(res_desc, tex_desc)

    return tex_obj, cu_arr


int2_t = np.dtype({"names": ["x", "y"], "formats": [np.int32, np.int32]})


def cupy_array_from_ptr(ptr, shape, dtype, owner_reference):
    size_bytes = np.prod(shape) * np.dtype(dtype).itemsize
    arr_cupy_mem = cuda.UnownedMemory(ptr, size_bytes, owner_reference)  # TODO device_id ?
    arr_cupy_memptr = cuda.MemoryPointer(arr_cupy_mem, offset=0)
    return cupy_ndarray(shape, dtype=dtype, memptr=arr_cupy_memptr)  # pylint: disable=E1123


def copy_to_cuarray_with_offset(dst_cuarray, src_array, offset_x=0, offset_y=0):
    """
    Copy a cupy.ndarray into a CUDA Array (matrix-like data structures that bakes textures).

    Copies a matrix ('height' rows of 'width' bytes each)
    from the memory area pointed to by 'src' to the CUDA array 'dst' starting at the upper left corner (wOffset, hOffset) [...]
    'spitch' is the width in memory in bytes of the 2D array pointed to by 'src', including any padding added to the end of each row.
      - 'wOffset' + 'width' must not exceed the width of the CUDA array 'dst'.
      - 'width' must not exceed 'spitch'.


    NB: for copies without offset, we can just use dst_cuarray.copy_from(src_cupyarray)
    """

    if isinstance(src_array, np.ndarray):
        memcpy_kind = cuda.runtime.memcpyHostToDevice
        src_array_ptr = src_array.ctypes.data
    else:
        memcpy_kind = cuda.runtime.memcpyDeviceToDevice
        src_array_ptr = src_array.data.ptr

    nbytes_per_item = np.dtype(src_array.dtype).itemsize
    h, w = src_array.shape

    w_offset = offset_x * nbytes_per_item
    h_offset = offset_y
    spitch = src_array.strides[-2]
    width = w * nbytes_per_item
    height = h

    cuda.runtime.memcpy2DToArray(
        dst_cuarray.ptr,  # intptr_t dst,
        w_offset,  # size_t wOffset : horizontal offset, probably in bytes
        h_offset,  # size_t hOffset : vertical offset, probably in pixels
        src_array_ptr,  # intptr_t src
        spitch,  # size_t spitch : source width in bytes (!) incl. padding (see cudaMallocPitch)
        width,  # size_t width : width of the matrix transfer, in bytes (!)
        height,  # size_t height : height of the matrix transfer, in pixels
        memcpy_kind,  # int kind: Type of transfer
    )


"""
def pycuda_to_cupy(arr_pycuda):
    arr_cupy_mem = cupy.cuda.UnownedMemory(arr_pycuda.ptr, arr_pycuda.size, arr_pycuda)
    arr_cupy_memptr = cupy.cuda.MemoryPointer(arr_cupy_mem, offset=0)
    return cupy.ndarray(arr_pycuda.shape, dtype=arr_pycuda.dtype, memptr=arr_cupy_memptr)  # pylint: disable=E1123


def cupy_to_pycuda(arr_cupy):
    return garray.empty(arr_cupy.shape, arr_cupy.dtype, gpudata=arr_cupy.data.ptr)
"""



import pytest
import numpy as np
from scipy.ndimage import shift, median_filter
from nabu.testutils import generate_tests_scenarios, get_data
from nabu.utils import clip_circle
from nabu.cuda.utils import __has_cupy__
from nabu.reconstruction.mlem import MLEMReconstructor, NabuMLEMReconstructor, has_corrct

__have_corrct__ = has_corrct(safe=True)
try:
    import astra

    __has_astra__ = True
except:
    __has_astra__ = False


@pytest.fixture(scope="class")
def bootstrap_corrct_mlem(request):
    cls = request.cls
    datafile = get_data("test_mlem.npz")
    cls.data_wvu = datafile["data_wvu"]
    cls.angles_rad = datafile["angles_rad"]
    cls.pixel_size_cm = datafile["pixel_size"] * 1e4  # pixel_size originally in um
    cls.true_cor = datafile["true_cor"]
    cls.mlem_cor_None_nosh = datafile["mlem_cor_None_nosh"]
    cls.mlem_cor_truecor_nosh = datafile["mlem_cor_truecor_nosh"]
    cls.mlem_cor_truecor_shifts_v0 = datafile["mlem_cor_truecor_shifts_v0"]
    cls.shifts_uv_v0 = datafile["shifts_uv_v0"]
    cls.shifts_uv = datafile["shifts_uv"]

    cls.tol = 1.3e-4


@pytest.mark.skipif(not (__has_cupy__ and __have_corrct__), reason="Need cupy and corrct for this test")
@pytest.mark.usefixtures("bootstrap_corrct_mlem")
class TestCorrctMLEMReconstructor:
    """These tests test the general MLEM reconstruction algorithm
    and the behavior of the reconstruction with respect to horizontal shifts.
    Only horizontal shifts are tested here because vertical shifts are handled outside
    the reconstruction object, but in the embedding reconstruction pipeline. See FullFieldReconstructor
    It is compared against a reference reconstruction generated with the `rec_mlem` function
    defined in the `generate_test_data.py` script.
    """

    def _rec_mlem(self, cor, shifts_uv, data_wvu, angles_rad):
        n_angles, n_z, n_x = data_wvu.shape

        mlem = MLEMReconstructor(
            (n_z, n_angles, n_x),
            angles_rad,
            shifts_uv=shifts_uv,
            cor=cor,
            n_iterations=50,
            extra_options={
                "centered_axis": True,
                "clip_outer_circle": True,
                "scale_factor": 1 / self.pixel_size_cm,
            },
        )
        rec_mlem = mlem.reconstruct(data_wvu.swapaxes(0, 1))
        return rec_mlem

    def test_simple_mlem_recons_cor_None_nosh(self):
        slice_index = 25
        rec = self._rec_mlem(None, None, self.data_wvu, self.angles_rad)[slice_index]
        delta = np.abs(rec - self.mlem_cor_None_nosh)
        assert np.max(delta) < self.tol

    def test_simple_mlem_recons_cor_truecor_nosh(self):
        slice_index = 25
        rec = self._rec_mlem(self.true_cor, None, self.data_wvu, self.angles_rad)[slice_index]
        delta = np.abs(rec - self.mlem_cor_truecor_nosh)
        assert np.max(delta) < 2.6e-4

    def test_compare_with_fbp(self):
        from nabu.reconstruction.fbp import Backprojector

        def _rec_fbp(cor, shifts_uv, data_wvu, angles_rad):
            n_angles, n_z, n_x = data_wvu.shape

            if shifts_uv is None:
                fbp = Backprojector(
                    (n_angles, n_x),
                    angles=angles_rad,
                    rot_center=cor,
                    halftomo=False,
                    padding_mode="edges",
                    extra_options={
                        "centered_axis": True,
                        "clip_outer_circle": True,
                        "scale_factor": 1 / self.pixel_size_cm,
                    },
                )
            else:
                fbp = Backprojector(
                    (n_angles, n_x),
                    angles=angles_rad,
                    rot_center=cor,
                    halftomo=False,
                    padding_mode="edges",
                    extra_options={
                        "centered_axis": True,
                        "clip_outer_circle": True,
                        "scale_factor": 1 / self.pixel_size_cm,  # convert um to cm
                        "axis_correction": shifts_uv[:, 0],
                    },
                )

            rec_fbp = np.zeros((n_z, n_x, n_x), "f")
            for i in range(n_z):
                rec_fbp[i] = fbp.fbp(data_wvu[:, i])

            return rec_fbp

        fbp = _rec_fbp(self.true_cor, None, self.data_wvu, self.angles_rad)[25]
        mlem = self._rec_mlem(self.true_cor, None, self.data_wvu, self.angles_rad)[25]
        delta = np.abs(fbp - mlem)
        assert (
            np.max(delta) < 400
        )  # These two should not be really equal. But the test should test that both algo FBP and MLEM behave similarly.

    def test_mlem_zeroshifts_equal_noshifts(self):
        shifts = np.zeros((len(self.angles_rad), 2))
        rec_nosh = self._rec_mlem(self.true_cor, None, self.data_wvu, self.angles_rad)
        rec_zerosh = self._rec_mlem(self.true_cor, shifts, self.data_wvu, self.angles_rad)
        delta = np.abs(rec_nosh - rec_zerosh)
        assert np.max(delta) < self.tol

    def test_mlem_recons_with_u_shifts(self):
        slice_index = 25
        rec = self._rec_mlem(self.true_cor, self.shifts_uv_v0, self.data_wvu, self.angles_rad)[slice_index]
        delta = np.abs(rec - self.mlem_cor_truecor_shifts_v0)
        assert np.max(delta) < self.tol


@pytest.fixture(scope="class")
def bootstrap_nabu_mlem(request):
    cls = request.cls
    cls.sino = get_data("mri_sino500.npz")["data"]
    cls.sino_halftomo = get_data("sino_halftomo.npz")["sinogram"].astype("f")
    cls.rot_center_halftomo = get_data("sino_halftomo.npz")["rot_center"]


def astra_mlem(sino, rec_shape, n_it):
    vol_geom = astra.create_vol_geom(rec_shape)
    proj_geom = astra.create_proj_geom("parallel", 1.0, sino.shape[-1], np.linspace(0, np.pi, sino.shape[0], False))

    rec_id = astra.data2d.create("-vol", vol_geom)
    sinogram_id = astra.data2d.create("-sino", proj_geom)
    astra.data2d.store(sinogram_id, sino)
    astra.data2d.store(rec_id, np.ones(rec_shape, "f"))  # !

    cfg = astra.astra_dict("EM_CUDA")
    cfg["ReconstructionDataId"] = rec_id
    cfg["ProjectionDataId"] = sinogram_id

    alg_id = astra.algorithm.create(cfg)
    astra.algorithm.run(alg_id, n_it)

    rec = astra.data2d.get(rec_id)

    astra.algorithm.delete(alg_id)
    astra.data2d.delete(rec_id)
    astra.data2d.delete(sinogram_id)

    return rec


def EM_reg_median(sino, P, PT, n_it, Lambda, eps=1e-6, debugperiod=None, domask=True):
    """
    Reference code for "regularized MLEM"
    """
    o = np.ones_like(sino)
    nangles = sino.shape[0]
    pto = PT(o) / nangles

    x = np.ones_like(pto)
    if domask:
        radius = (sino.shape[1] - 1) / 2
        a, b = (radius, radius)
        nx, ny = x.shape
        yr, xr = np.ogrid[-a : nx - a, -b : ny - b]
        mask = xr * xr + yr * yr > radius * radius
        if Lambda:
            x[mask] = 0

    for k in range(n_it):
        proj = P(x)
        M = np.abs(proj) > eps
        proj_inv = np.copy(sino)
        proj_inv[M] = sino[M] / proj[M]
        med_x = median_filter(x, size=(3, 3))
        # med_x[med_x < eps] = eps
        dPen = Lambda * (x - med_x) * med_x / (med_x * med_x + eps)
        if debugperiod is not None and (k % debugperiod) == 0:
            print("Iteration={0}|{1}, Min={2:g}, Max={3:g}".format(k, n_it, dPen.min(), dPen.max()))
        oinv = 1.0 / (pto + dPen)
        oinv[oinv < 0] = 0.0
        x *= PT(proj_inv) / nangles * oinv
    if domask:
        mask = xr * xr + yr * yr > (radius - 2) * (radius - 2)
        x[mask] = 0
    return x


@pytest.mark.skipif(not (__has_cupy__ and __has_astra__), reason="Need astra for this test")
@pytest.mark.usefixtures("bootstrap_nabu_mlem")
class TestNabuMLEM:

    def test_mlem_simple(self):
        n_it = 50

        sino = self.sino
        mlem = NabuMLEMReconstructor(sino.shape)
        rec = mlem.reconstruct(sino, n_iterations=n_it).get()
        ref = astra_mlem(sino, mlem.slice_shape, n_it)

        err_max = np.max(np.abs(rec - ref))
        tol = 4.5e-2
        assert err_max < tol, f"MLEM error is too high: {err_max} > {tol}"

    def test_mlem_rotcenter(self):
        subsampling = 2
        n_it = 50
        # has to be int for this test. Don't choose a too high value to avoid cropping the sinogram content
        cor_offset = 10

        sino0 = self.sino[::subsampling]

        mlem0 = NabuMLEMReconstructor(sino0.shape)
        rec0 = mlem0.reconstruct(sino0, n_iterations=n_it).get()
        rec0_crop = rec0[cor_offset // 2 : -cor_offset // 2, cor_offset // 2 : -cor_offset // 2]

        sino = sino0[:, cor_offset:]
        rot_center = (sino0.shape[-1] - 1) / 2 - cor_offset
        mlem = NabuMLEMReconstructor(sino.shape, rot_center=rot_center)
        rec = mlem.reconstruct(sino, n_iterations=n_it).get()

        err_max = np.max(np.abs(rec - rec0_crop))
        tol = 0.5  # proj/backproj are not really translation-invariant after all
        assert err_max < tol

    def test_mlem_translations(self):
        n_iterations = 50

        mlem0 = NabuMLEMReconstructor(self.sino.shape)
        rec0 = mlem0.reconstruct(self.sino, n_iterations=n_iterations).get()

        # translations are in pixel units
        transl_min = -11
        transl_max = 20.5
        rng_seed = 10
        rng = np.random.Generator(np.random.PCG64(seed=rng_seed))
        x_translations = rng.uniform(low=transl_min, high=transl_max, size=self.sino.shape[0]).astype("f")

        def _sino_horizonal_shifts(sino):
            sino_shifted = sino.copy()
            for i in range(sino.shape[0]):
                sino_shifted[i] = shift(sino[i], x_translations[i], order=2)
            return sino_shifted

        sino_shifted = _sino_horizonal_shifts(self.sino)

        mlem = NabuMLEMReconstructor(self.sino.shape, extra_options={"axis_correction": x_translations})
        rec = mlem.reconstruct(sino_shifted, n_iterations=n_iterations).get()

        err_max = np.max(np.abs(rec - rec0))
        tol = 4.0  # small errors accumulate when doing iterative. Still quite good looking
        assert err_max < tol, f"Wrong reconstruction with x-translations, using seed={rng_seed}"

    def test_mlem_clip_circle(self):
        sino = self.sino[:, 100:-100]  # crop the sinogram a little bit to see edges effect for clip_circle
        n_it = 30
        err_max = lambda x, y: np.max(np.abs(x - y))
        tol = 6e-2

        mlem = NabuMLEMReconstructor(sino.shape, extra_options={"clip_outer_circle": True})
        rec = mlem.reconstruct(sino, n_iterations=n_it).get()
        mlem_noclip = NabuMLEMReconstructor(sino.shape, extra_options={"clip_outer_circle": False})
        rec_noclip = mlem_noclip.reconstruct(sino, n_iterations=n_it).get()

        ref = astra_mlem(sino, mlem.slice_shape, n_it)

        assert not (np.allclose(rec, rec_noclip)), "MLEM+clip is supposed to yield a different result than MLEM_plain"
        assert (
            err_max(rec, clip_circle(rec_noclip)) < 1e-7
        ), "MLEM+clip is supposed to yield the same result as clip(MLEM_plain)"

        ref = clip_circle(ref)
        err_max = np.max(np.abs(rec - ref))
        assert err_max < tol, f"MLEM + clip_circle error is too high: {err_max} > {tol}"

    def test_mlem_halftomo(self):
        sino = self.sino_halftomo
        n_it = 50

        mlem = NabuMLEMReconstructor(
            sino.shape,
            angles=np.linspace(0, 2 * np.pi, sino.shape[0], False),
            rot_center=self.rot_center_halftomo,
            halftomo=True,
            extra_options={"clip_outer_circle": True},
        )

        rec = mlem.reconstruct(sino, n_iterations=n_it).get()

        # When reconstructing local-tomography data, it's better to use a larger support for backprojection,
        # to attenuate the "cup artefact". The data is cropped at the end of the process
        mlem2 = NabuMLEMReconstructor(
            sino.shape,
            angles=np.linspace(0, 2 * np.pi, sino.shape[0], False),
            rot_center=self.rot_center_halftomo,
            halftomo=True,
            extra_options={"clip_outer_circle": False},
            slice_shape=(1501, 1501),
        )

        dy = (mlem2.slice_shape[-1] - mlem.slice_shape[-1]) // 2
        rec2 = mlem.reconstruct(sino).get()[dy:-dy, dy:-dy]

    def test_mlem_median_reg(self):
        n_it = 30
        sino = self.sino

        params_to_test = {
            "regularization_weight": [0, 1.4],
            "clip_outer_circle": [True, False],
        }
        # See issue #678
        err_tol = {(False, False): 4e-4, (False, True): 5e-2, (True, True): 4.4, (True, False): 7.5}

        for config in generate_tests_scenarios(params_to_test):
            weight = config["regularization_weight"]
            do_clip_circle = config["clip_outer_circle"]

            # Test plain regularization (no mask)
            mlem = NabuMLEMReconstructor(
                sino.shape,
                extra_options={
                    "clip_outer_circle": do_clip_circle,
                    "reg_weight": weight,
                },
            )
            rec = mlem.reconstruct(sino, n_iterations=n_it, regularization_weight=weight).get()
            ref = EM_reg_median(
                sino,
                mlem.proj.projection,
                mlem.backproj.backproj,
                n_it,
                Lambda=weight,
                domask=do_clip_circle,
            )

            err_max = np.max(np.abs(rec - ref))

            tol = err_tol[(weight > 0, do_clip_circle)]
            assert err_max < tol, f"regularized-MLEM error is too high for config={config}: {err_max} > {tol}"



import pytest
from nabu.testutils import get_dummy_nxtomo_info
from nabu.resources.dataset_analyzer import analyze_dataset


@pytest.fixture(scope="class")
def bootstrap_nx(request):
    cls = request.cls
    cls.nx_fname, cls.data_desc, cls.image_key, cls.projs_vals, cls.darks_vals, cls.flats1_vals, cls.flats2_vals = (
        get_dummy_nxtomo_info()
    )


@pytest.mark.usefixtures("bootstrap_nx")
class TestNXDataset:

    def test_exclude_projs_angular_range(self):
        dataset_info_with_all_projs = analyze_dataset(self.nx_fname)

        # Test exclude angular range - angles min and max in degrees
        angular_ranges_to_test = [(0, 15), (5, 6), (50, 58.5)]
        for angular_range in angular_ranges_to_test:
            angle_min, angle_max = angular_range
            dataset_info = analyze_dataset(
                self.nx_fname,
                extra_options={"exclude_projections": {"type": "angular_range", "range": [angle_min, angle_max]}},
            )
            excluded_projs_indices = dataset_info.get_excluded_projections_indices()
            # Check that get_excluded_projections_indices() angles are correct
            for excluded_proj_index in excluded_projs_indices:
                frame_angle_deg = dataset_info.dataset_scanner.frames[excluded_proj_index].rotation_angle
                assert angle_min <= frame_angle_deg <= angle_max

            assert set(dataset_info_with_all_projs.projections.keys()) - set(dataset_info.projections.keys()) == set(
                excluded_projs_indices
            )
